/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <string.h>
#include <math.h>

#include "kiba-akamaru.h"

static GList *objects = NULL;

/**
Helper Functions
**/

void kiba_akamaru_do_for_all_objects (KibaAkamaruModel *model,
				      KibaAkamaruFunc   func,
				      gpointer          data) {
	GList *l;
	
	for (l = objects; l != NULL; l = l->next) {
		KibaAkamaruObject *object = l->data;
		
		func (model, object, data);
	}
}

static void kiba_akamaru_object_del_spring (KibaAkamaruModel  *model,
					    KibaAkamaruObject *object,
					    gpointer           data) {
	if (object->spring == NULL)
		return;
	
	akamaru_model_delete_spring (&model->model, object->spring); 
	object->spring = NULL;
}

static void kiba_akamaru_object_add_spring (KibaAkamaruModel  *model,
					    KibaAkamaruObject *object,
					    gpointer           data) {
	if (object->spring != NULL)
		return;
	
	object->spring = akamaru_model_add_spring (&model->model,
		model->anchor->object, object->object, 0);
}

static void kiba_akamaru_object_del_wobbly_spring (KibaAkamaruModel *model,
						   KibaAkamaruObject *object,
						   gpointer           data) {
	GList *l;
	
	for (l = object->wobbly_spring; l != NULL; l = l->next) {
		AkamaruSpring *wobbly_spring = l->data;
	
		akamaru_model_delete_spring (&model->model, wobbly_spring);
	}
	
	object->wobbly_spring = NULL;
}

static void kiba_akamaru_add_wobbly_spring (KibaAkamaruModel *model) {
	AkamaruObject *last_object = NULL;
	GList         *l;
	
	for (l = objects; l != NULL; l = l->next) {
		KibaAkamaruObject *object = l->data;
	
		if (model->model_flag == AKAMARU_SPRING) {
			GList *li;
			
			for (li = objects; li != NULL; li = li->next) {
				KibaAkamaruObject *buf_object = li->data;
				
				if (buf_object == object)
					continue;
				
				object->wobbly_spring = g_list_append (object->wobbly_spring,
					akamaru_model_add_spring (&model->model, object->object,
						buf_object->object, model->wobbly_spring));
			}
		}
		else if (last_object != NULL)
			object->wobbly_spring = g_list_append (object->wobbly_spring,
					akamaru_model_add_spring (&model->model, object->object, last_object, -model->wobbly_spring*5));
	
			last_object = object->object;
	}
}

static void kiba_akamaru_object_del_rope_strings (KibaAkamaruModel *model,
						  KibaAkamaruObject *object,
						  gpointer           data) {
	if (object->rope_string == NULL)
		return;
	
	akamaru_model_delete_string (&model->model, object->rope_string);
	object->rope_string = NULL;
}

static void kiba_akamaru_add_rope_strings (KibaAkamaruModel *model) {
	KibaAkamaruObject *buf_object = NULL;
	GList *l;
	
	for (l = objects; l != NULL; l = l->next) {
		KibaAkamaruObject *object = l->data;
	
		if (buf_object != NULL && object->rope_string == NULL) {
			float space;
	
			space = buf_object->object->radius + object->object->radius;
	
			object->rope_string = akamaru_model_add_string (&model->model,
				object->object, buf_object->object,
				(model->rope_string_length < 1.0f ?
				space : space * model->rope_string_length/10.0f));
		}
	
		buf_object = object;
	}
}

static void add_spacer (AkamaruObject *object,
			void          *data) {
	KibaAkamaruObject *closure_object = data;
	
	/* Skip the anchor object. */
	if (object != closure_object->model->anchor->object && object != closure_object->object)
		akamaru_model_add_spacer (&closure_object->model->model,
			closure_object->object, object, object->radius + closure_object->object->radius);
}
	
static void kiba_akamaru_reload_rope_string (KibaAkamaruModel *model) {
	if (model->model_flag != AKAMARU_ROPE)
		return;
	
	kiba_akamaru_do_for_all_objects (model, kiba_akamaru_object_del_rope_strings, NULL);
	kiba_akamaru_add_rope_strings (model);
}

static void kiba_akamaru_reload_wobbly_spring (KibaAkamaruModel *model) {
	if (model->model_flag != AKAMARU_WOBBLY && model->model_flag != AKAMARU_SPRING)
		return;
	
	kiba_akamaru_do_for_all_objects (model, kiba_akamaru_object_del_wobbly_spring, NULL);
	kiba_akamaru_add_wobbly_spring(model);
}

static gboolean position_is_occupied (gint   x,
				      gint   y,
				      float radius) {
	GList *l;
	
	for (l = objects; l != NULL; l = l->next) {
		KibaAkamaruObject *obj = l->data;
		float dx, dy, distance;
		
		dx = (float) x - obj->object->position.x;
		dy = (float) y - obj->object->position.y;
		distance = sqrt(dx*dx + dy*dy);
		
		if (distance < obj->object->radius + radius)
			return TRUE;
	}
	
	return FALSE;
}

/**
Properties
**/

static const char *akamaru_model_data[] = {
	"dock", "rope", "wobbly", "spring"
};

void kiba_akamaru_set_model_from_string (KibaAkamaruModel *model,
					 const char       *string) {
	gint i;
	
	if (model->model_flag == AKAMARU_ROPE)
		kiba_akamaru_do_for_all_objects (model, kiba_akamaru_object_del_rope_strings, NULL);
	else if (model->model_flag == AKAMARU_WOBBLY || model->model_flag == AKAMARU_SPRING)
		kiba_akamaru_do_for_all_objects (model, kiba_akamaru_object_del_wobbly_spring, NULL);
	
	for (i = 0; i < G_N_ELEMENTS(akamaru_model_data); i++) {
		if (strcmp (string, akamaru_model_data[i]) == 0) {
			model->model_flag = i;
			break;
		}
	}
	
	if (model->model_flag == AKAMARU_ROPE)
		kiba_akamaru_add_rope_strings (model);
	else if (model->model_flag == AKAMARU_WOBBLY || model->model_flag == AKAMARU_SPRING)
		kiba_akamaru_add_wobbly_spring(model);
}

void kiba_akamaru_set_k (KibaAkamaruModel *model,
			 float            k) {
	model->k = k;
	
	if (model->active)
		model->model.k = model->k;
}

void kiba_akamaru_set_friction (KibaAkamaruModel *model,
				float            friction) {
	model->friction = friction;
	
	if (model->active)
		model->model.k = model->friction;
}

void kiba_akamaru_set_elasticity (KibaAkamaruModel *model,
				  float            elasticity) {
	model->elasticity = elasticity;
	
	if (model->active)
		model->model.k = model->elasticity;
}

void kiba_akamaru_set_wobbly_spring (KibaAkamaruModel *model,
				     float            wobbly_spring) {
	model->wobbly_spring = wobbly_spring;
	kiba_akamaru_reload_wobbly_spring (model);
}

void kiba_akamaru_set_inertia (KibaAkamaruModel *model,
			       float            inertia) {
	model->inertia = inertia;
}

void kiba_akamaru_set_gravity (KibaAkamaruModel *model,
			       float            gravity) {
	model->gravity = gravity;
}

void kiba_akamaru_set_throw_gravity (KibaAkamaruModel *model,
				     float            throw_gravity) {
	model->throw_gravity = throw_gravity;
}

void kiba_akamaru_set_rope_string_length (KibaAkamaruModel *model,
					  float            rope_string_length) {
	model->rope_string_length = rope_string_length;
	kiba_akamaru_reload_rope_string (model);
}

void kiba_akamaru_model_enable (KibaAkamaruModel *model,
				gboolean          enable) {
	model->active = enable;
}

/**
Model
**/

gboolean kiba_akamaru_is_enabled (KibaAkamaruModel *model) {
	return model->active;
}

void kiba_akamaru_unset_mouse_anchor (KibaAkamaruModel *model) {
	model->mouse_anchor->object  = NULL;
}

void kiba_akamaru_set_mouse_anchor (KibaAkamaruModel  *model,
				    KibaAkamaruObject *object,
				    float             x,
				    float             y) {
	if (!object)
		return;
	model->mouse_anchor->x = x + object->object->radius;
	model->mouse_anchor->y = y + object->object->radius;
	model->mouse_anchor->object = object->object;
}

void kiba_akamaru_set_anchor_position (KibaAkamaruModel *model,
				       int               x,
				       int               y) {
	model->anchor->x = x;
	model->anchor->y = y;
}

void kiba_akamaru_update_mouse_anchor (KibaAkamaruModel *model,
				       int               new_x,
				       int               new_y,
				       float            icon_size,
				       GdkRectangle      screen_geometry) {
	model->mouse_anchor->x = CLAMP (new_x, (int) icon_size / 2,
			screen_geometry.x + screen_geometry.width - (int) icon_size /2 );
	model->mouse_anchor->y = CLAMP (new_y, (int) icon_size / 2,
			screen_geometry.y + screen_geometry.height - (int) icon_size / 2);
}

void kiba_akamaru_get_mouse_anchor_position (KibaAkamaruModel *model,
					     int              *anchor_x,
					     int              *anchor_y) {
	*anchor_x = model->mouse_anchor->x;
	*anchor_y = model->mouse_anchor->y;
}

void kiba_akamaru_model_set_size (KibaAkamaruModel *model,
				  GdkRectangle      screen_geometry) {
	akamaru_model_delete_polygon (&model->model, model->desktop_polygon);
	model->desktop_polygon = akamaru_model_add_enclosing_rectangle (&model->model,
		screen_geometry.x,  screen_geometry.y,
		screen_geometry.x + screen_geometry.width,
		screen_geometry.y + screen_geometry.height);
	
/*
	FIXME: engine is acting strange when changing the anchor points
	physics->anchor->x       = screen_geometry.x;
	physics->anchor->y       = screen_geometry.y;
	physics->mouse_anchor->x = screen_geometry.x;
	physics->mouse_anchor->y = screen_geometry.y;
*/
}

void kiba_akamaru_model_step (KibaAkamaruModel *model) {
	gint i;
	
	for (i = 0; i < 6; i++)
		akamaru_model_step (&model->model, 0.01/model->inertia);
}

void kiba_akamaru_model_free (KibaAkamaruModel *model) {
	if (kiba_akamaru_is_enabled(model)) {
		if (model->anchor != NULL)
			akamaru_model_delete_anchor (&model->model, model->anchor);
		
		if (model->mouse_anchor != NULL)
			akamaru_model_delete_anchor (&model->model, model->mouse_anchor);
	}
	
	if (model->anchor != NULL)
		akamaru_model_delete_anchor (&model->model, model->anchor);
	
	if (model->mouse_anchor != NULL)
		akamaru_model_delete_anchor (&model->model, model->mouse_anchor);
	
	akamaru_model_fini (&model->model);
	
	g_free (model);
}

KibaAkamaruModel* kiba_akamaru_model_new (GdkRectangle screen_geometry) {
	KibaAkamaruModel *model;
	
	model = g_new0 (KibaAkamaruModel, 1);
	
	model->constrain_iterations = 1;
	model->elasticity = 0.7f;
	model->friction = 10;
	model->k = 0.6f;
	model->rope_string_length = 13.5f;
	model->throw_gravity = 10;
	model->gravity = 60;
	model->inertia = 2.0f;
	model->wobbly_spring = 5000;
	model->model_flag = AKAMARU_DOCK;
	model->geometry = screen_geometry;
	model->active = FALSE;
	
	model->anchor = NULL;
	model->mouse_anchor = NULL;
	
	akamaru_model_init (&model->model);
	
	model->model.constrain_iterations = model->constrain_iterations;
	model->model.elasticity = model->elasticity;
	model->model.friction = model->friction;
	model->model.k = model->k;
	
	/* add a polygon around the desktop so that the objects reflect at the borders of the desktop */
	model->desktop_polygon = akamaru_model_add_enclosing_rectangle (&model->model,
		model->geometry.x,  model->geometry.y,
		model->geometry.x + model->geometry.width,
		model->geometry.y + model->geometry.height);
	
	/* add the anchor object */
	model->anchor = akamaru_model_add_anchor (&model->model,
				akamaru_model_add_object (&model->model, 0, 0, 0, NULL),
				model->geometry.x, model->geometry.y);
	model->mouse_anchor = akamaru_model_add_anchor (&model->model,
				NULL, model->geometry.x, model->geometry.y);
	
	return model;
}


/**
Objects
**/

void kiba_akamaru_reload_spacers (KibaAkamaruModel *model) {
	GList *l;
	
	akamaru_model_delete_spacers (&model->model);
	
	for (l = objects; l != NULL; l = l->next) {
		KibaAkamaruObject *object = l->data;
		
		akamaru_model_for_each_object (&model->model, add_spacer, object);
	}
}

void kiba_akamaru_object_throw (KibaAkamaruModel   *model,
				KibaAkamaruObject *object,
				int                dx,
				int                dy) {
	model->mouse_anchor->object  = NULL;
	
	object->object->previous_position.x -= dx / model->throw_gravity;
	object->object->previous_position.y -= dy / model->throw_gravity;
}

void kiba_akamaru_object_bounce (KibaAkamaruModel   *model,
				 KibaAkamaruObject *object,
				 GdkRectangle       bounce_point) {
	model->mouse_anchor->object  = NULL;
	
	object->object->previous_position.x += bounce_point.x;
	object->object->previous_position.y += bounce_point.y;
}

void kiba_akamaru_object_set_gravity_point (KibaAkamaruModel   *model,
					    KibaAkamaruObject *object,
					    GdkRectangle       gravity_normal) {
	object->object->gravity.x = gravity_normal.x * model->gravity;
	object->object->gravity.y = gravity_normal.y * model->gravity;
}

void kiba_akamaru_object_set_position (KibaAkamaruObject *object,
				       float             x,
				       float             y) {
	object->object->previous_position.x = object->object->position.x;
	object->object->previous_position.y = object->object->position.y;
	object->object->position.x = x;
	object->object->position.y = y;
}

void kiba_akamaru_object_free (KibaAkamaruModel  *model,
			       KibaAkamaruObject *object) {
	GList *l;
	
	objects = g_list_remove (objects, object);
	
	if (object->spring != NULL) {
		akamaru_model_delete_spring (&model->model, object->spring);
		object->spring = NULL;
	}
	
	if (object->rope_string != NULL) {
		akamaru_model_delete_string (&model->model, object->rope_string);
		object->rope_string = NULL;
	}
	
	if (object->object != NULL) {
		akamaru_model_delete_object (&model->model, object->object);
		object->object = NULL;
	}
	
	if (object->wobbly_spring) {
		for (l = object->wobbly_spring; l != NULL; l = l->next) {
			AkamaruSpring *spring = l->data;
	
			akamaru_model_delete_spring (&model->model, spring);
			spring = NULL;
		}
		g_list_free (object->wobbly_spring);
	}
	
	g_free (object);
	object = NULL;
	
	if (model->model_flag == AKAMARU_ROPE) {
		kiba_akamaru_do_for_all_objects (model, kiba_akamaru_object_del_rope_strings, NULL);
		kiba_akamaru_add_rope_strings (model);
	}
	else if (model->model_flag == AKAMARU_WOBBLY || model->model_flag == AKAMARU_SPRING) {
		kiba_akamaru_do_for_all_objects (model, kiba_akamaru_object_del_wobbly_spring, NULL);
		kiba_akamaru_add_wobbly_spring(model);
	}
	
	/* reload spacer */
	kiba_akamaru_reload_spacers (model);
}

KibaAkamaruObject* kiba_akamaru_add_object (KibaAkamaruModel *model,
					    gint              width,
					    gint              height,
					    float            radius) {
	KibaAkamaruObject *object;
	gint x, y;
	
	object = g_new0 (KibaAkamaruObject, 1);
	
	x = g_random_int_range (1, width);
	y = g_random_int_range (1, height);
/*
	i = 0;
	while (position_is_occupied (x, y, radius)) {
		if (i > 100) {
			fprintf (stderr, "Cant find a valid position for Akamaru Object\n");
			return NULL;
		}
		x = g_random_int_range (1, width);
		y = g_random_int_range (1, height);
		i++;
	}
*/
	
	object->object = akamaru_model_add_object (&model->model, x, y, 12, NULL);
	object->model = model;
	object->object->friction = model->model.friction;
	object->object->radius = radius;
	/* Add spacer (for Object reflection) */
	akamaru_model_for_each_object (&model->model, add_spacer, object);
	
	object->spring = akamaru_model_add_spring (&model->model,
					model->anchor->object,
					object->object, 0);
	
	objects = g_list_append (objects, object);
	
	/* reload rope strings */
	if (model->model_flag == AKAMARU_ROPE) {
		kiba_akamaru_do_for_all_objects (model, kiba_akamaru_object_del_rope_strings, NULL);
		kiba_akamaru_add_rope_strings (model);
	}
	/* reload wobbly/spring springs */
	else if (model->model_flag == AKAMARU_WOBBLY || model->model_flag == AKAMARU_SPRING) {
		kiba_akamaru_do_for_all_objects (model, kiba_akamaru_object_del_wobbly_spring, NULL);
		kiba_akamaru_add_wobbly_spring(model);
	}
	
	return object;
}

