/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <string.h>
#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <libgnomeui/libgnomeui.h>

#include <kiba-win.h>
#include <kiba-debug.h>
#include <kiba-prefs.h>
#include <kiba-object.h>
#include <kiba-globals.h>
#include <kiba-inline-pixbufs.h>

#include "stack-icon.h"

#define KIBA_STACK_ICON_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_STACK_ICON, KibaStackIconPrivate))

/**
Class Info:
	Each KibaStackIcon instance represents one File in the Filesystem. it tries to find the standart application for this file, but also allows setting of custom applications.
	The struct contains a cairo_surface_t surface buffer displaying a custom icon, thumbnail, mime icon or the application icon.
	KibaStackIcon's are usually grouped together in a KibaStack instance.
**/

typedef struct _KibaStackIconPrivate KibaStackIconPrivate;

struct _KibaStackIconPrivate {
	gint dump;
};

enum {
	ACTIVATE_EVENT = 0,
	DEACTIVATE_EVENT,
	SIG_STACK_ICON_LAST
};

static guint stack_icon_signals[SIG_STACK_ICON_LAST] = { 0 };

static void     kiba_stack_icon_init              (KibaStackIcon      *self);
static void     kiba_stack_icon_class_init        (KibaStackIconClass *klass);

static gpointer kiba_stack_icon_parent_class = NULL;
static GType    kiba_stack_icon_type_id = 0;

static void     kiba_stack_icon_class_intern_init (gpointer klass)
{
  kiba_stack_icon_parent_class = g_type_class_peek_parent (klass); 
  kiba_stack_icon_class_init ((KibaStackIconClass*) klass); 
}

GType
kiba_stack_icon_get_type (void)
{
  return kiba_stack_icon_type_id;
}

void
kiba_stack_icon_register_type (GTypeModule *type_module)
{
  const GTypeInfo g_define_type_info = {
    sizeof (KibaStackIconClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) kiba_stack_icon_class_intern_init,
    (GClassFinalizeFunc) NULL,
    NULL,   /* class_data */
    sizeof (KibaStackIcon),
    0,      /* n_preallocs */
    (GInstanceInitFunc) kiba_stack_icon_init, 
    NULL    /* value_table */
  };
  kiba_stack_icon_type_id = g_type_module_register_type (type_module,
					      G_TYPE_OBJECT,
					      "KibaStackIcon",
					      &g_define_type_info, 0);
}

/**
Helper Functions
**/

static void kiba_stack_icon_draw_surface (cairo_t          *cr,
					  cairo_surface_t  *surface,
					  gint              width,
					  gint              height,
					  void             *data) {
	KibaStackIcon *stack_icon = data;
	KibaStackIconPrivate *priv;
	GdkPixbuf *pixbuf = NULL;
	const char *icon;
	gint size, icon_width, icon_height;
	
	size = MAX(width,height);
	priv = KIBA_STACK_ICON_GET_PRIVATE(stack_icon);
	
	if (stack_icon->app)
		icon = gnome_vfs_mime_application_get_icon (stack_icon->app);
	else
		icon = NULL;
	
	/* first try the custom icon */
	if (stack_icon->custom_icon)
		kiba_load_icon(stack_icon->custom_icon, size, &pixbuf);
	/* now the mime/thumbnail icon */
	if (!pixbuf && stack_icon->mime_icon)
		kiba_load_icon(stack_icon->mime_icon, size, &pixbuf);
	/* now the application icon */
	if (!pixbuf && icon)
		kiba_load_icon(icon, size, &pixbuf);
	/* no icon found, using fallback :/ */
	if (!pixbuf)
		pixbuf = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
	
	icon_width  = gdk_pixbuf_get_width (pixbuf);
	icon_height = gdk_pixbuf_get_height (pixbuf);
	
	gdk_cairo_set_source_pixbuf (cr, pixbuf,
		(width - icon_width)/2,
		(height - icon_height)/2);
	
	cairo_paint (cr);
	g_object_unref (pixbuf);
}

/**
(re)load the surface buffer for the KibaStackIcon instance
**/
void kiba_stack_icon_update_surface (KibaStackIcon *icon) {
	KibaObject *object;
	cairo_t *cr;
	gboolean has_dock_object;
	
	has_dock_object = KIBA_IS_STACK(icon->stack);
	if (has_dock_object) {
		object =  KIBA_OBJECT(icon->stack->widget);
		
		icon->surface->width   = object->size.width * object->scale_x;
		icon->surface->height  = object->size.height * object->scale_x;
	}
	
	if (icon->type == KIBA_STACK_ICON_TYPE_MIME_FILE ||
	   (!icon->custom_icon && icon->type == KIBA_STACK_ICON_TYPE_DIRECTORY)) {
		if (icon->surface->surface) {
			cairo_surface_destroy (icon->surface->surface);
			icon->surface->surface = NULL;
		}
		icon->surface->surface = kiba_stack_get_mime_surface (icon->stack,
						      icon->mime_type,
						      icon->mime_icon);
	}
	if (icon->type != KIBA_STACK_ICON_TYPE_MIME_FILE || !icon->surface->surface) {
		cr = kiba_win_cairo_create (kiba_win_get_default());
		cairo_update_surface (cr, icon->surface,
				kiba_stack_icon_draw_surface, icon);
		cairo_destroy (cr);
	}
	
	if (has_dock_object && kiba_stack_icon_get_state(icon) == KIBA_STACK_STATE_ACTIVE) {
		object->surface_data->surface = cairo_surface_reference (icon->surface->surface);
		kiba_object_queue_draw (object);
	}
}

/**
set all url related variables of stack_icon (mime_type, icon, exec, name and type)
**/
static gboolean kiba_stack_icon_load_from_url (KibaStackIcon *stack_icon,
					       const char    *url,
					       GError       **error) {
	KibaStackIconPrivate *priv;
	GQuark  error_quark;
	GnomeVFSResult result;
	GnomeVFSFileInfo info;
	GnomeThumbnailFactory *thumb_factory;
	char *utf8_uri = NULL;
	
	priv = KIBA_STACK_ICON_GET_PRIVATE(stack_icon);
	error_quark = g_quark_from_string ("kiba-dock");
	
	stack_icon->mime_type = gnome_vfs_get_mime_type (url);
	if (!stack_icon->mime_type) {
		g_set_error (error, error_quark, 1, "Failed to get MIME Type from url '%s'", url);
		return FALSE;
	}
	
	if (!g_utf8_validate (url, -1, NULL)) {
		utf8_uri = g_locale_to_utf8 (url, -1, NULL, NULL, NULL);
	}
	if (!utf8_uri) {
		utf8_uri = g_strdup(url);
	}
	
	stack_icon->url = g_strdup(url);
	stack_icon->type = -1;
	stack_icon->name = g_strdup(g_strrstr(utf8_uri, "/")+1);
	
	thumb_factory = stack_icon->stack->stack_plugin->thumb_factory;
	
	/* its a desktop file, we gonna create a special launcher icon */
	if (g_str_has_suffix (url, ".desktop")) {
		static const char desktop_entry_group[] = "Desktop Entry";
		
		GKeyFile *file;
		GError *error = NULL;
		char *icon_name;
		
		file = g_key_file_new ();
		if (g_key_file_load_from_file (file, url, 0, &error)) {
			stack_icon->mime_icon = g_key_file_get_value (file, desktop_entry_group, "Icon", &error);
			g_clear_error (&error);
			error = NULL;
			stack_icon->exec = g_key_file_get_value (file, desktop_entry_group, "Exec", &error);
			g_clear_error (&error);
			error = NULL;
			if (!stack_icon->exec) {
				char *buf_key;
				
				buf_key = g_key_file_get_value (file, desktop_entry_group, "URL", &error);
				g_clear_error (&error);
				error = NULL;
				if (!buf_key) {
					buf_key = g_key_file_get_value (file, desktop_entry_group, "Path", &error);
					g_clear_error (&error);
					error = NULL;
				}
				if (buf_key) {
					stack_icon->exec = g_strdup_printf ("%s %s", file_manager_global, buf_key);
					g_free (buf_key);
				}
			}
			g_key_file_free (file);
		}
		
		icon_name = g_strndup (stack_icon->name,
				strlen(stack_icon->name) - strlen(".desktop"));
		g_free (stack_icon->name);
		stack_icon->name = icon_name;
		if (!stack_icon->mime_icon &&
		    gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (), icon_name, 48, 0)) {
			stack_icon->mime_icon = icon_name;
		}
		
		stack_icon->type = KIBA_STACK_ICON_TYPE_DESKTOP;
	}
	/* get the icon from the file info */
	else if (gnome_vfs_get_file_info (url, &info, GNOME_VFS_FILE_INFO_DEFAULT) == GNOME_VFS_OK) {
		GnomeVFSURI *vfs_uri;
		char *uri_string = NULL;
		char *escaped_uri;
		
		escaped_uri = gnome_vfs_make_uri_from_input (url);
		vfs_uri = gnome_vfs_uri_new (escaped_uri);
		g_free (escaped_uri);
		if (vfs_uri != NULL) {
			uri_string = gnome_vfs_uri_to_string (vfs_uri, GNOME_VFS_URI_HIDE_NONE);
			gnome_vfs_uri_unref (vfs_uri);
		}
		
		/* try to find a thumbnail */
		if (uri_string && gnome_thumbnail_factory_can_thumbnail (thumb_factory,
						   uri_string,
						   stack_icon->mime_type,
						   info.mtime)) {
			stack_icon->type = KIBA_STACK_ICON_TYPE_THUMB_FILE;
			stack_icon->mime_icon = gnome_thumbnail_factory_lookup (thumb_factory,
								uri_string, info.mtime);
			if (!stack_icon->mime_icon) {
				kiba_stack_load_thumb_for_icon (stack_icon);
			}
		}
		
		if (!stack_icon->mime_icon) {
			stack_icon->type = KIBA_STACK_ICON_TYPE_MIME_FILE;
			stack_icon->mime_icon = gnome_icon_lookup (gtk_icon_theme_get_default(),
							NULL, url, NULL,
							&info, stack_icon->mime_type,
							GNOME_ICON_LOOKUP_FLAGS_NONE, NULL);
		}
	}
	if (g_file_test(url, G_FILE_TEST_IS_DIR)) {
		stack_icon->type = KIBA_STACK_ICON_TYPE_DIRECTORY;
	}
	
	stack_icon->app  = gnome_vfs_mime_get_default_application_for_uri (url, stack_icon->mime_type);
	/* get the exec */
	if (!stack_icon->exec) {
		const char *custom_command_line;
		char *vfs_uri = NULL;
		char *validate;
		
		vfs_uri = gnome_vfs_make_uri_from_shell_arg (utf8_uri);
		
		/* this character is not escaped by the gnome vfs escape function :/ */
		if (g_strrstr (vfs_uri, "'")) {
			char *buf = vfs_uri;
			
			vfs_uri = gnome_vfs_escape_set (buf, "'");
			g_free (buf);
		}
		custom_command_line = kiba_stack_get_custom_command_for_mime_type (stack_icon->stack->stack_plugin, stack_icon->mime_type);
		if (custom_command_line) {
			stack_icon->exec = g_strdup_printf ("%s %s", custom_command_line, vfs_uri);
		}
		else if (stack_icon->app) {
			const char *command_line;
			
			command_line = gnome_vfs_mime_application_get_exec (stack_icon->app);
			if (command_line)
				stack_icon->exec = g_strdup_printf ("%s %s", command_line, vfs_uri);
		}
		
		if (!stack_icon->mime_icon) {
			const char *app_icon;
			
			app_icon = gnome_vfs_mime_application_get_icon (stack_icon->app);
			if (app_icon)
				stack_icon->mime_icon = g_strdup(app_icon);
		}
		
		g_free (vfs_uri);
	}
	
	return TRUE;
}

/**
use url as file for icon
**/
gboolean kiba_stack_icon_set_url (KibaStackIcon *icon,
			          const char    *url) {
	GError *error = NULL;
	
	g_return_if_fail (KIBA_IS_STACK_ICON(icon));
	
	if (!kiba_stack_icon_load_from_url(icon, url, &error)) {
		DEBUG_ERROR ("Failed to get KibaStackIcon from url '%s'\n", url);
		if (error != NULL) {
			DEBUG_ERROR ("%s\n", error->message);
			g_free (error);
		}
		return FALSE;
	}
	return TRUE;
}

/**
Properties
**/

/**
sets wether the icon is active, only one icon can be active,
it provides the surface for the dock object
**/
void kiba_stack_icon_set_state (KibaStackIcon      *icon,
				KibaStackIconState  state) {
	g_return_if_fail (KIBA_IS_STACK_ICON(icon));
	
	if (state == icon->state)
		return;
	
	icon->state = state;
	
	if (state == KIBA_STACK_STATE_ACTIVE) {
		if (KIBA_IS_STACK(icon->stack)) {
			KibaObject *object;
			
			object = KIBA_OBJECT(icon->stack->widget);
			
			if (icon->name)
				kiba_object_set_title (object, icon->name);
			
			if (object->surface_data->surface)
				cairo_surface_destroy (object->surface_data->surface);
			object->surface_data->surface =
				cairo_surface_reference (icon->surface->surface);
			kiba_object_queue_draw (object);
		}
		
		g_signal_emit (icon, stack_icon_signals[ACTIVATE_EVENT], 0);
	}
	else
		g_signal_emit (icon, stack_icon_signals[DEACTIVATE_EVENT], 0);
}

/**
return wether the icon is active
**/
KibaStackIconState kiba_stack_icon_get_state (KibaStackIcon *icon) {
	g_return_val_if_fail (KIBA_IS_STACK_ICON(icon), KIBA_STACK_STATE_INACTIVE);
	
	return icon->state;
}

/**
Class Functions
**/

static void kiba_stack_icon_class_init (KibaStackIconClass *class) {
	
	g_type_class_add_private (class, sizeof (KibaStackIconPrivate));
	
	stack_icon_signals[ACTIVATE_EVENT] =
		g_signal_new ("activate-event",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaStackIconClass, activate),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
	stack_icon_signals[DEACTIVATE_EVENT] =
		g_signal_new ("deactivate-event",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaStackIconClass, deactivate),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
}

static void kiba_stack_icon_init (KibaStackIcon *icon) {
	icon->state = KIBA_STACK_STATE_INACTIVE;
	
	icon->stack = NULL;
	
	icon->mime_icon = NULL;
	icon->custom_icon = NULL;
	icon->exec = NULL;
	
	icon->surface = g_new0 (CairoSurfaceData, 1);
	icon->surface->surface = NULL;
	icon->surface->width   = -1;
	icon->surface->height  = -1;
}

KibaStackIcon* kiba_stack_icon_new (KibaStack *stack) {
	KibaStackIcon *icon;
	
	if (!kiba_stack_icon_get_type())
		kiba_stack_icon_register_type (G_TYPE_MODULE(stack->stack_plugin->plugin));
	icon = g_object_new (KIBA_TYPE_STACK_ICON, NULL);
	icon->stack = stack;
	
	return icon;
}

void kiba_stack_icon_destroy (KibaStackIcon *icon) {
	g_return_if_fail (KIBA_IS_STACK_ICON(icon));
	
	if (icon->url)
		g_free (icon->url);
	
	cairo_surface_destroy (icon->surface->surface);
	g_free (icon->surface);
	
	g_object_unref (icon);
}
