/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <string.h>
#include <libintl.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <libgnomevfs/gnome-vfs-utils.h>

#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#include "stack.h"

#define _(string)  dgettext (GETTEXT_PACKAGE, string)

/**
Plugin Info:
	The stack plugin is responsable for adding files and directories
	to objects on the dock.
	The intention is, having files that are frequently used just a few clicks away on the dock,
	but you also can use the stack icon-view to browse through the complete filesystem.
**/

/**
Helper Functions
**/

static void kiba_stack_drag_dest_func (GdkDragContext *drag_context,
				       const char     *drag_data,
				       gpointer        user_data,
				       char          **action_ret,
				       char          **icon_ret) {
	static const char *file_prefix = "file://";
	
	char *string;
	char *unescaped;
	gboolean valid_file_found;
	
	char data[strlen(drag_data)];
	sprintf(data, "%s", drag_data);
	g_strstrip (data);
	
	valid_file_found = FALSE;
	while ((string = g_strrstr (data, "\n")) != NULL) {
		g_strstrip (string);
		
		if (g_str_has_prefix(string, file_prefix)) {
			unescaped = gnome_vfs_unescape_string (string + strlen(file_prefix), NULL);
		}
		else {
			unescaped = gnome_vfs_unescape_string (string, NULL);
		}
		
		if (!g_file_test(unescaped, G_FILE_TEST_EXISTS)) {
			string[0] = '\0';
			continue;
		}
		
		valid_file_found = TRUE;
		
		string[0] = '\0';
	}
	if (g_str_has_prefix(drag_data, file_prefix)) {
		unescaped = gnome_vfs_unescape_string (drag_data + strlen(file_prefix), NULL);
	}
	else {
		unescaped = gnome_vfs_unescape_string (drag_data, NULL);
	}
	
	if (g_file_test(unescaped, G_FILE_TEST_EXISTS))
		valid_file_found = TRUE;
	
	if (!valid_file_found)
		return;
	
	*action_ret = g_strdup(_("Create Stack"));
	*icon_ret   = g_strdup("file-manager");
}

/**
the dock received a drag,
we gonna create a new KibaStack instance if there are any valid files in the drag data
**/
static gboolean kiba_stack_drag_receive_func (GdkDragContext *drag_context,
					      const char     *drag_data,
					      gpointer        user_data) {
	static const char *file_prefix = "file://";
	static const char *stackfile_prefix = "stackfile";
	static const char *stackfile_suffix = ".stack";
	
	KibaStackPlugin *plugin = user_data;
	KibaStack *stack;
	GList *files = NULL;
	GList *l;
	char *string;
	char *dnd_data;
	char *filename;
	char *unescaped;
	const char *homedir;
	FILE *stack_file;
	
	dnd_data = g_strdup(drag_data);
	
	while ((string = g_strrstr (dnd_data, "\n")) != NULL) {
		g_strstrip (string);
		
		if (g_str_has_prefix(string, file_prefix)) {
			unescaped = gnome_vfs_unescape_string (string + strlen(file_prefix), NULL);
		}
		else {
			unescaped = gnome_vfs_unescape_string (string, NULL);
		}
		filename = unescaped;
		
		if (!g_file_test(unescaped, G_FILE_TEST_EXISTS)) {
			string[0] = '\0';
			continue;
		}
		
		files = g_list_append (files, g_strdup(filename));
		g_free (unescaped);
		
		string[0] = '\0';
	}
	g_strstrip (dnd_data);
	
	if (g_str_has_prefix(dnd_data, file_prefix)) {
		unescaped = gnome_vfs_unescape_string (dnd_data + strlen(file_prefix), NULL);
	}
	else {
		unescaped = gnome_vfs_unescape_string (dnd_data, NULL);
	}
	filename = unescaped;
	if (g_file_test(unescaped, G_FILE_TEST_EXISTS)) {
		files = g_list_append (files, g_strdup(filename));
	}
	g_free (unescaped);
	g_free (dnd_data);
	
	if (!files)
		return FALSE;
	
	homedir = g_get_home_dir ();
	
	char stack_dir[strlen(homedir)+strlen(STACK_DIR)];
	sprintf (stack_dir, "%s" STACK_DIR, homedir);
	
	filename = g_strdup_printf ("%s-0%s", stackfile_prefix, stackfile_suffix);
	
	char uri[strlen(stack_dir) + strlen(filename) + 3];
	sprintf (uri, "%s/%s", stack_dir, filename);
	
	gint i = 0;
	while (g_file_test(uri, G_FILE_TEST_EXISTS)) {
		g_free (filename);
		filename = g_strdup_printf ("%s-%i%s", stackfile_prefix, i+1, stackfile_suffix);
		sprintf (uri, "%s/%s", stack_dir, filename);
		i++;
	}
	
	stack_file = fopen (uri, "w");
	if (stack_file == NULL)
		return FALSE;
	for (l = files; l != NULL; l = l->next) {
		fprintf (stack_file, "%s\n", (char*) l->data);
	}
	fclose (stack_file);
	
	stack = kiba_stack_new (plugin, uri);
	plugin->stacks = g_list_append (plugin->stacks, stack);
	g_free (filename);
	
	return FALSE;
}

static void populate_dialog_response_cb (GtkWidget *widget,
					 gint       arg1,
					 gpointer   user_data) {
	static const char *stackfile_prefix = "stackfile";
	static const char *stackfile_suffix = ".stack";
	
	KibaStackPlugin *plugin = user_data;
	KibaStack *stack;
	GSList *files, *l;
	FILE *stack_file;
	const char *homedir;
	char *stack_filename;
	
	if (arg1 != GTK_RESPONSE_ACCEPT) {
		gtk_widget_destroy (widget);
		return;
	}
	
	homedir = g_get_home_dir ();
	
	char stack_dir[strlen(homedir)+strlen(STACK_DIR)];
	sprintf (stack_dir, "%s" STACK_DIR, homedir);
	
	stack_filename = g_strdup_printf ("%s-0%s", stackfile_prefix, stackfile_suffix);
	
	char uri[strlen(stack_dir) + strlen(stack_filename) + 3];
	sprintf (uri, "%s/%s", stack_dir, stack_filename);
	
	gint i = 0;
	while (g_file_test(uri, G_FILE_TEST_EXISTS)) {
		g_free (stack_filename);
		stack_filename = g_strdup_printf ("%s-%i%s", stackfile_prefix, i+1, stackfile_suffix);
		sprintf (uri, "%s/%s", stack_dir, stack_filename);
		i++;
	}
	g_free (stack_filename);
	
	stack_file = fopen (uri, "w");
	if (stack_file == NULL)
		return;
	
	files = gtk_file_chooser_get_filenames (GTK_FILE_CHOOSER(widget));
	for (l = files; l != NULL; l = l->next) {
		fprintf (stack_file, "%s\n", (char*) l->data);
	}
	fclose (stack_file);
	
	stack = kiba_stack_new (plugin, uri);
	plugin->stacks = g_list_append (plugin->stacks, stack);
	
	gtk_widget_destroy (widget);
}

static void kiba_stack_populate_dock (KibaDock *dock,
				      gpointer  data) {
	KibaStackPlugin *plugin = data;
	GtkWidget *dialog;
	
	dialog = gtk_file_chooser_dialog_new (_("Choose Files for a Stack"), NULL,
		GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
	gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER (dialog), TRUE);
	
	gtk_widget_show (dialog);
	g_signal_connect (dialog, "response",
		G_CALLBACK (populate_dialog_response_cb), plugin);
}

static void icon_theme_changed_cb (GtkIconTheme *icon_theme,
				   gpointer      user_data) {
	KibaStackPlugin *plugin = user_data;
	GList *l;
	
	for (l = plugin->stacks; l != NULL; l = l->next) {
		KibaStack *stack = l->data;
		GList *li;
		
		if (!KIBA_IS_OBJECT(stack->widget))
			continue;
		
		for (li = stack->icons; li != NULL; li = li->next) {
			KibaStackIcon *icon= li->data;
			
			kiba_stack_icon_update_surface (icon);
		}
	}
}

/**
KibaPrefs Callbacks
**/

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaStackPlugin *plugin = data;
	KibaDock *dock;
	GList *l;
	float size_scale;
	
	dock = kiba_dock_get();
	
	size_scale = g_ascii_strtod((char*) value, NULL);
	
	for (l = plugin->stacks; l != NULL; l = l->next) {
		KibaStack *stack = l->data;
		GList *li;
		
		if (!KIBA_IS_OBJECT(stack->widget))
			continue;
		
		kiba_object_set_requisition_factor (KIBA_OBJECT(stack->widget),
			size_scale, size_scale);
		KIBA_OBJECT(stack->widget)->surface_data->width = KIBA_OBJECT(stack->widget)->size.width;
		KIBA_OBJECT(stack->widget)->surface_data->height = KIBA_OBJECT(stack->widget)->size.height;
		
		for (li = stack->icons; li != NULL; li = li->next) {
			KibaStackIcon *icon= li->data;
			
			icon->surface->width  = KIBA_OBJECT(stack->widget)->size.width;
			icon->surface->height = KIBA_OBJECT(stack->widget)->size.height;
			kiba_stack_icon_update_surface (icon);
		}
	}
	gtk_widget_queue_resize (GTK_WIDGET(dock));
}

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaStackPlugin *plugin = data;
	GList *l;
	
	for (l = plugin->stacks; l != NULL; l = l->next) {
		KibaStack *stack = l->data;
		GList *li;
		
		kiba_object_resize (KIBA_OBJECT(stack->widget));
		KIBA_OBJECT(stack->widget)->surface_data->width  = KIBA_OBJECT(stack->widget)->size.width;
		KIBA_OBJECT(stack->widget)->surface_data->height = KIBA_OBJECT(stack->widget)->size.height;
		
		for (li = stack->icons; li != NULL; li = li->next) {
			KibaStackIcon *icon= li->data;
			
			icon->surface->width  = KIBA_OBJECT(stack->widget)->size.width;
			icon->surface->height = KIBA_OBJECT(stack->widget)->size.height;
			kiba_stack_icon_update_surface (icon);
		}
	}
}

static void stack_view_columns_cb (const char *option_id,
				   gpointer    value,
				   gpointer    data) {
	KibaStackPlugin *plugin = data;
	KibaStack *stack;
	GList *l;
	
	plugin->view_columns = GPOINTER_TO_INT(value);
	
	stack = plugin->view_stack;
	if (stack) {
		kiba_stack_toggle_icon_view(stack);
		kiba_stack_toggle_icon_view(stack);
	}
}

static void stack_view_rows_cb (const char *option_id,
				gpointer    value,
				gpointer    data) {
	KibaStackPlugin *plugin = data;
	KibaStack *stack;
	GList *l;
	
	plugin->view_rows = GPOINTER_TO_INT(value);
	
	stack = plugin->view_stack;
	if (stack) {
		kiba_stack_toggle_icon_view(stack);
		kiba_stack_toggle_icon_view(stack);
	}
}

static void stack_view_slide_step_cb (const char *option_id,
				      gpointer    value,
				      gpointer    data) {
	KibaStackPlugin *plugin = data;
	GList *l;
	
	plugin->slide_step = GPOINTER_TO_INT(value);
	
	if (plugin->view_win) {
		KibaIconView *icon_view;
		
		icon_view = KIBA_ICON_VIEW(plugin->view_win->icon_view);
		icon_view->slide_step = plugin->slide_step;
	}
}

static void stack_view_open_dirs_cb (const char *option_id,
				     gpointer    value,
				     gpointer    data) {
	KibaStackPlugin *plugin = data;
	
	plugin->open_dirs = GPOINTER_TO_INT(value);
}

static void stack_view_icon_size_cb (const char *option_id,
				     gpointer    value,
				     gpointer    data) {
	KibaStackPlugin *plugin = data;
	KibaStack *stack;
	GList *l;
	gint size;
	
	plugin->view_icon_size = GPOINTER_TO_INT(value);
	
	stack = plugin->view_stack;
	if (stack) {
		kiba_stack_toggle_icon_view(stack);
		kiba_stack_toggle_icon_view(stack);
	}
}

static void view_text_color_cb (const char *option_id,
				gpointer    value,
				gpointer    data) {
	KibaStackPlugin *plugin = data;
	KibaStack *stack;
	
	plugin->text_color = (char*) value;
	
	stack = plugin->view_stack;
	if (stack) {
		kiba_stack_toggle_icon_view(stack);
		kiba_stack_toggle_icon_view(stack);
	}
}

static void view_text_style_cb (const char *option_id,
				gpointer    value,
				gpointer    data) {
	KibaStackPlugin *plugin = data;
	KibaStack *stack;
	
	plugin->text_italic = GPOINTER_TO_INT(value);
	
	stack = plugin->view_stack;
	if (stack) {
		kiba_stack_toggle_icon_view(stack);
		kiba_stack_toggle_icon_view(stack);
	}
}

static void stack_object_alpha_cb (const char *option_id,
				   gpointer    value,
				   gpointer    data) {
	KibaStackPlugin *plugin = data;
	GList *l;
	float alpha;
	
	alpha = g_ascii_strtod((char*) value, NULL);
	
	alpha = MAX(alpha, -alpha);
	alpha = MIN(alpha, 1.0);
	
	for (l = plugin->stacks; l != NULL; l = l->next) {
		KibaStack *stack = l->data;
		
		if (KIBA_IS_OBJECT(stack->widget))
			kiba_object_set_alpha (KIBA_OBJECT(stack->widget), alpha);
	}
}

static void stack_custom_mime_app_cb (const char *option_id,
				      gpointer    value,
				      gpointer    data) {
	KibaStackPlugin *plugin = data;
	GSList *l;
	
	for (l = (GSList*) value; l != NULL; l = l->next) {
		char *data = l->data;
		char *mime_type;
		char *command;
		char *buf;
		gint i;
		
		if (!g_strrstr(data, ":"))
			continue;
		
		i = 0;
		while (data[i] != ':')
			i++;
		
		mime_type = g_strndup(data, i);
		buf = data+i+1;
		command = g_strdup (buf);
		kiba_stack_set_custom_mime_app (plugin, mime_type, command);
		g_free (mime_type);
		g_free (command);
	}
}

/**
Plugin Functions
**/

static void stack_plugin_unload (KibaPlugin *plugin) {
	KibaStackPlugin *stack_plugin;
	GList *l;
	char custom_mime_apps[2000];
	
	stack_plugin = g_object_get_data (G_OBJECT(plugin), STACK_PLUGIN_NAME);
	g_object_set_data (G_OBJECT(plugin),
		STACK_PLUGIN_NAME, NULL);
	
	kiba_prefs_remove_callback_by_data (plugin);
	kiba_win_remove_drag_func (kiba_win_get_default(),
			     kiba_stack_drag_receive_func);
	kiba_dock_remove_populate_func (kiba_dock_get(),
		(KibaDockPopulateFunc) kiba_stack_populate_dock);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(gtk_icon_theme_get_default()),
		G_CALLBACK(icon_theme_changed_cb), stack_plugin);
	
	sprintf (custom_mime_apps, "[");
	for (l = stack_plugin->custom_mime_apps; l != NULL; l = l->next) {
		KibaStackCustomMimeApp *app = l->data;
		
		sprintf (custom_mime_apps, "%s%s:%s", custom_mime_apps, app->mime_type, app->command);
		if (l->next)
			sprintf (custom_mime_apps, "%s,", custom_mime_apps);
	}
	sprintf (custom_mime_apps, "%s]", custom_mime_apps);
	kiba_prefs_write_to_config_file ("stack_custom_mime_apps", "list", custom_mime_apps);
	
	for (l = stack_plugin->surface_pool; l != NULL; l = stack_plugin->surface_pool) {
		KibaStackMimeSurface *mime_surface = l->data;
		
		stack_plugin->surface_pool = g_list_remove (stack_plugin->surface_pool, l->data);
		
		cairo_surface_destroy (mime_surface->surface);
		g_free (mime_surface);
	}
	g_list_free (stack_plugin->surface_pool);
	
	for (l = stack_plugin->stacks; l != NULL; l = stack_plugin->stacks) {
		KibaStack *stack = l->data;
		FILE *stack_file;
		
		if (!KIBA_IS_STACK(stack))
			continue;
		
		if (!stack->icons) {
			if (g_file_test (stack->uri, G_FILE_TEST_EXISTS))
				g_remove (stack->uri);
			kiba_stack_destroy (stack);
			stack_plugin->stacks = g_list_remove (stack_plugin->stacks, stack);
			continue;
		}
		
		/* save stack */
		stack_file = fopen (stack->uri, "w");
		if (stack_file != NULL) {
			GList *li;
			
			for (li = stack->icons; li != NULL; li = li->next) {
				KibaStackIcon *icon = li->data;
				
				fprintf (stack_file, "%s\n", icon->url);
			}
			
			fclose (stack_file);
		}
		
		stack_plugin->stacks = g_list_remove (stack_plugin->stacks, stack);
		kiba_stack_destroy (stack);
	}
	g_list_free (stack_plugin->stacks);
	
	g_free (stack_plugin);
}

static gboolean stack_plugin_load (KibaPlugin *plugin) {
	KibaStackPlugin *stack_plugin;
	KibaDock *dock;
	GError *error = NULL;
	const char *homedir;
	char *file_name;
	GDir *dir;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
	
	homedir = g_get_home_dir ();
	
	char stack_dir[strlen(homedir)+strlen(STACK_DIR)];
	sprintf (stack_dir, "%s" STACK_DIR, homedir);
	if (!g_file_test (stack_dir, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir (stack_dir, 7664) != 0) {
			DEBUG_ERROR ("Failed to Create %s\n", stack_dir);
			return FALSE;
		}
	}
	
	if (!gnome_vfs_initialized () && !gnome_vfs_init ()) {
		DEBUG_ERROR ("Can not initialise gnome-vfs.\n");
		return FALSE;
	}
	
	stack_plugin = g_new0 (KibaStackPlugin, 1);
	stack_plugin->plugin = plugin;
	stack_plugin->custom_mime_apps = NULL;
	stack_plugin->stacks = NULL;
	stack_plugin->surface_pool = NULL;
	stack_plugin->thumb_factory =
		gnome_thumbnail_factory_new (GNOME_THUMBNAIL_SIZE_NORMAL);
	stack_plugin->view_win = NULL;
	stack_plugin->view_stack = NULL;
	stack_plugin->view_type = KIBA_STACK_TYPE_NONE;
	stack_plugin->text_color = g_strdup("#000000");
	stack_plugin->text_italic = FALSE;
	stack_plugin->view_icon = NULL;
	
	kiba_prefs_add_callback ("dock_object_size",
		dock_object_size_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_custom_mime_apps",
		stack_custom_mime_app_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_view_txt_color",
		view_text_color_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_view_text_italic",
		view_text_style_cb, stack_plugin, TRUE);
	
	/* add stack files from homedir */
	dir = g_dir_open (stack_dir, 0, &error);
	if (error != NULL) {
		DEBUG_ERROR ("Failed to open '%s': %s\n",
			stack_dir, error->message);
		g_error_free (error);
		return FALSE;
	}
	
	while ((file_name = (char*) g_dir_read_name (dir)) != NULL) {
		KibaStack *stack;
		
		if (!g_str_has_suffix (file_name, ".stack"))
			continue;
		
		char uri[strlen(stack_dir) + strlen(file_name) + 1];
		sprintf (uri, "%s/%s", stack_dir, file_name);
		
		stack = kiba_stack_new (stack_plugin, uri);
		stack_plugin->stacks = g_list_append (stack_plugin->stacks, stack);
	}
	g_dir_close (dir);
	
	kiba_prefs_add_callback ("stack_size_factor",
		object_size_factor_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_alpha",
		stack_object_alpha_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_view_columns",
		stack_view_columns_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_view_rows",
		stack_view_rows_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_slide_step",
		stack_view_slide_step_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_open_dirs",
		stack_view_open_dirs_cb, stack_plugin, TRUE);
	kiba_prefs_add_callback ("stack_view_icon_size",
		stack_view_icon_size_cb, stack_plugin, TRUE);
	
	g_signal_connect (G_OBJECT (gtk_icon_theme_get_default()), "changed",
			G_CALLBACK (icon_theme_changed_cb), stack_plugin);
	
	kiba_win_add_drag_func (kiba_win_get_default(),
			kiba_stack_drag_receive_func,
			kiba_stack_drag_dest_func,
			stack_plugin);
	kiba_dock_add_populate_func (kiba_dock_get(),
		(KibaDockPopulateFunc) kiba_stack_populate_dock,
		"file-manager",
		_("Add Stack"),
		(gpointer) stack_plugin);
	
	g_object_set_data (G_OBJECT(plugin),
		STACK_PLUGIN_NAME, stack_plugin);
	
	return TRUE;
}

static PrefMember stack_options[] = {
  { "stack_alpha", "1.0", KIBA_DOUBLE },
  { "stack_custom_mime_apps", "[]", KIBA_LIST },
  { "stack_size_factor", "1.0", KIBA_DOUBLE },
  { "stack_view_rows", "4", KIBA_INT },
  { "stack_view_columns", "4", KIBA_INT },
  { "stack_view_icon_size", "64", KIBA_INT },
  { "stack_view_txt_color", "#FFFFFF", KIBA_STRING },
  { "stack_view_text_italic", "FALSE", KIBA_BOOLEAN },
  { "stack_slide_step", "10", KIBA_INT },
  { "stack_open_dirs", "TRUE", KIBA_BOOLEAN },
};

static KibaPluginVtable vtable = {
	STACK_PLUGIN_NAME,		/* id/name        */
	"Stack Plugin for Kiba-Dock",/* summary        */
	NULL,				/* description    */	

	"Daniel Bessler <danielb@kiba-dock.org>",
					/* author         */
	"http://www.kiba-dock.org",	/* homepage       */

	stack_plugin_load,		/* load            */
	stack_plugin_unload,		/* unload          */
};

KIBA_INIT_PLUGIN (vtable, stack_options);
