/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <string.h>
#include <libintl.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <libgnomevfs/gnome-vfs-xfer.h>
#include <libgnomevfs/gnome-vfs-async-ops.h>

#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-object.h>
#include <kiba-menu-items.h>
#include <kiba-info-win.h>
#include <kiba-icon-view.h>
#include <kiba-image.h>
#include <kiba-screen.h>
#include <kiba-utils.h>
#include <kiba-app-chooser.h>

#include "stack.h"

#define _(string)  dgettext (GETTEXT_PACKAGE, string)

#define MAX_OBJECTS           VIEW_COLUMNS*VIEW_MAX_ROWS
#define STACK_APPLICATION_DIR "/usr/share/applications"

/**
Class Info:
	Each stack instance represents one Dock object, it contains multiple KibaStackIcon's.
	The currently selected KibaStackIcon provides the icon for the object on the dock.
	KibaStack uses a KibaIconViewWin to show the user a grid containing all KibaStackIcon and to show the content of the activated directory icon.
	a new KibaStack can be created by dragging files to the dock,
	the stack can be filled by dragging files or directories to the dock object or to the icon view.
**/

enum {
	ICON_ADDED_EVENT = 0,
	ICON_REMOVED_EVENT,
	SIG_KIBA_STACK_LAST
};

static KibaStackIcon *stack_dnd_icon = NULL;

static guint kiba_stack_signals[SIG_KIBA_STACK_LAST] = { 0 };

static void kiba_stack_create_group_for_icon (KibaStack    *stack,
					      KibaStackIcon *icon);
static GtkWidget* create_icon_for_uri (KibaStack   *stack,
				       const char  *uri);
static gboolean create_icon_view_for_directory_icon (KibaStackIcon *directory_icon);

static void     kiba_stack_init              (KibaStack      *self);
static void     kiba_stack_class_init        (KibaStackClass *klass);

static gpointer kiba_stack_parent_class = NULL;
static GType    kiba_stack_type_id = 0;

static void     kiba_stack_class_intern_init (gpointer klass)
{
  kiba_stack_parent_class = g_type_class_peek_parent (klass); 
  kiba_stack_class_init ((KibaStackClass*) klass); 
}

GType
kiba_stack_get_type (void)
{
  return kiba_stack_type_id;
}

static void
kiba_stack_register_type (GTypeModule *type_module)
{
  const GTypeInfo g_define_type_info = {
    sizeof (KibaStackClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) kiba_stack_class_intern_init,
    (GClassFinalizeFunc) NULL,
    NULL,   /* class_data */
    sizeof (KibaStack),
    0,      /* n_preallocs */
    (GInstanceInitFunc) kiba_stack_init, 
    NULL    /* value_table */
  };
  kiba_stack_type_id = g_type_module_register_type (type_module,
					      G_TYPE_OBJECT,
					      "KibaStack",
					      &g_define_type_info, 0);
}

/**
Helper Functions
**/

/**
Sort helper function, to sort the icons by their name
**/
static gint kiba_stack_sort_icons_by_name (gconstpointer apointer,
					   gconstpointer bpointer) {
	const KibaStackIcon *aicon = apointer;
	const KibaStackIcon *bicon = bpointer;
	
	return kiba_utils_sort_string_list (aicon->name, bicon->name);
}

/**
Select the given KibaStackIcon (switching dock object surface)
**/
static void kiba_stack_select_icon (KibaStack *stack,
				    KibaStackIcon *icon) {
	kiba_object_switch_surface (KIBA_OBJECT(stack->widget));
	
	kiba_stack_icon_set_state (KIBA_STACK_ICON(stack->icons->data),
			KIBA_STACK_STATE_INACTIVE);
	while (stack->icons->data != icon)
		kiba_utils_glist_scroll (&stack->icons, GDK_SCROLL_DOWN);
	kiba_stack_icon_set_state (icon, KIBA_STACK_STATE_ACTIVE);
}

/**
Returns wether a file was allready added to the stack
**/
static gboolean kiba_stack_contains_file (KibaStack  *stack,
					  const char *path) {
	GList *l;
	
	for (l = stack->icons; l != NULL; l = l->next) {
		KibaStackIcon *icon = l->data;
		
		if (strcmp (icon->url, path) == 0)
			return TRUE;
	}
	
	return FALSE;
}

static gboolean kiba_stack_remove_icon_cb (GtkWidget *widget,
					   gpointer   data) {
	KibaStackIcon *icon = data;
	KibaStack *stack = icon->stack;
	
	if (!KIBA_IS_STACK_ICON(icon))
		return TRUE;
	
	if (g_list_find(icon->stack->icons, icon))
		kiba_stack_remove (stack, icon);
	kiba_stack_icon_destroy (icon);
	
	return TRUE;
}

static gboolean kiba_stack_remove_cb (GtkWidget *widget,
				      gpointer   data) {
	KibaStack *stack = data;
	GList *l;
	
	for (l = stack->icons; l != NULL; l = stack->icons) {
		kiba_stack_remove_icon_cb (widget, l->data);
	}
}

/**
Application selected callback from KibaAppChooser,
we gonna set the custom mime app for the type of the icon
**/
static void kiba_stack_icon_app_selected (KibaAppChooser *chooser,
					  const char     *desktop,
					  gpointer        user_data) {
	static const char desktop_entry_group[] = "Desktop Entry";
	
	KibaStackIcon *icon = (KibaStackIcon*) user_data;
	KibaStackCustomMimeApp *app;
	GKeyFile *file;
	GList *l;
	GError *error = NULL;
	char *exec;
	
	file = g_key_file_new ();
	if (!g_key_file_load_from_file (file, desktop, 0, &error)) {
		g_key_file_free (file);
		return;
	}
	
	exec = g_key_file_get_value (file, desktop_entry_group, "Exec", &error);
	g_clear_error (&error);
	error = NULL;
	g_key_file_free (file);
	if (!exec)
		return;
	
	/* change exec of all KibaStackIcon's */
	for (l = icon->stack->stack_plugin->stacks; l != NULL; l = l->next) {
		KibaStack *stack = l->data;
		GList *li;
		
		for (li = stack->icons; li != NULL; li = li->next) {
			KibaStackIcon *check_icon = li->data;
			char *vfs_uri;
			
			if (strcmp (check_icon->mime_type, icon->mime_type) != 0)
				continue;
			
			vfs_uri = gnome_vfs_make_uri_from_shell_arg (check_icon->url);
			g_free (check_icon->exec);
			check_icon->exec = g_strdup_printf ("%s %s", exec, vfs_uri);
			g_free (vfs_uri);
		}
	}
	if (icon->stack->stack_plugin->view_win) {
		KibaIconView *icon_view;
		
		icon_view = KIBA_ICON_VIEW(icon->stack->stack_plugin->view_win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaStackIcon *check_icon;
			char *vfs_uri;
			
			check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
			
			if (strcmp (check_icon->mime_type, icon->mime_type) != 0)
				continue;
			
			vfs_uri = gnome_vfs_make_uri_from_shell_arg (check_icon->url);
			g_free (check_icon->exec);
			check_icon->exec = g_strdup_printf ("%s %s", exec, vfs_uri);
			g_free (vfs_uri);
		}
	}
	
	/* launch the icon with the new mime app */
	kiba_launch_programm (icon->exec);
	
	/* check if the mime type is already associated */
	for (l = icon->stack->stack_plugin->custom_mime_apps; l != NULL; l = l->next) {
		app = l->data;
		
		if (strcmp (app->mime_type, icon->mime_type) == 0) {
			g_free (app->command);
			app->command = g_strdup (exec);
			return;
		}
	}
	
	/* create new custom mime app */
	app = g_new0 (KibaStackCustomMimeApp, 1);
	app->command = g_strdup (exec);
	app->mime_type = g_strdup (icon->mime_type);
	g_free (exec);
	
	icon->stack->stack_plugin->custom_mime_apps =
		g_list_append (icon->stack->stack_plugin->custom_mime_apps, app);
}

/**
the "Set Mime Application" menu item was pressed,
we gonna create and popup a KibaAppChooser instance
**/
static void kiba_stack_set_mime_app_cb (GtkWidget *widget,
					gpointer   user_data) {
	KibaAppChooser *chooser;
	KibaStackIcon *icon = user_data;
	
	chooser = kiba_app_chooser_new(STACK_APPLICATION_DIR);

// 	KIBA_INFO_WIN(chooser->view_win->win)->gradient_start = icon->stack->stack_plugin->view_rgb1;
// 	KIBA_INFO_WIN(chooser->view_win->win)->gradient_mid   = icon->stack->stack_plugin->view_rgb2;
// 	KIBA_INFO_WIN(chooser->view_win->win)->gradient_end   = icon->stack->stack_plugin->view_rgb3;
// 	KIBA_INFO_WIN(chooser->view_win->win)->border_color   = icon->stack->stack_plugin->view_border_rgb;
	gtk_widget_show (chooser->view_win->win);
	gdk_window_set_keep_above (chooser->view_win->win->window, TRUE);
	
	g_signal_connect (G_OBJECT(chooser), "app-activate",
		G_CALLBACK(kiba_stack_icon_app_selected), user_data);
}

/**
a drag was received on the custom icon view or on the dock object,
we gonna add new KibaStackIcon's to the stack for the dragged files
**/
static gboolean kiba_stack_drag_receive_cb (KibaStack        *stack,
					    GdkDragContext   *drag_context,
					    GtkSelectionData *selection_data) {
	static const char *file_prefix = "file://";
	
	KibaStackPlugin *plugin = stack->stack_plugin;
	KibaStackIcon *icon = NULL;
	char *string;
	const char *homedir;
	char *filename;
	char *unescaped;
	gboolean file_added;
	GList *l;
	
	homedir = g_get_home_dir ();
	
	char stack_dir[strlen(homedir)+strlen(STACK_DIR)];
	sprintf (stack_dir, "%s" STACK_DIR, homedir);
	
	char data[strlen((char*) selection_data->data)];
	sprintf(data, "%s", (char*) selection_data->data);
	g_strstrip (data);
	
	file_added = FALSE;
	
	/* walk through the drag data (the items are seperated with "\n") */
	while ((string = g_strrstr (data, "\n")) != NULL) {
		g_strstrip (string);
		
		/* get the filename */
		if (g_str_has_prefix(string, file_prefix)) {
			unescaped = gnome_vfs_unescape_string (string + strlen(file_prefix), NULL);
		}
		else {
			unescaped = gnome_vfs_unescape_string (string, NULL);
		}
		filename = unescaped;
		
		/* check if its a valid file */
		if (!g_file_test(filename, G_FILE_TEST_EXISTS)) {
			string[0] = '\0';
			continue;
		}
		
		/* skip if the file was allready added */
		if (kiba_stack_contains_file (stack, filename)) {
			string[0] = '\0';
			continue;
		}
		
		/* create, add and select a new KibaStackIcon instance */
		icon = kiba_stack_icon_new(stack);
		if (!kiba_stack_icon_set_url (icon, filename)) {
			kiba_stack_icon_destroy (icon);
			continue;
		}
		g_free (unescaped);
		
		kiba_stack_add (stack, icon);
		kiba_stack_select_icon (stack, icon);
		
		file_added = TRUE;
		
		string[0] = '\0';
	}
	g_strstrip (data);
	/* same for the last item */
	if (g_str_has_prefix(data, file_prefix)) {
		unescaped = gnome_vfs_unescape_string (data + strlen(file_prefix), NULL);
	}
	else {
		unescaped = gnome_vfs_unescape_string (data, NULL);
	}
	filename = unescaped;
	
	if (g_file_test(filename, G_FILE_TEST_EXISTS) && !kiba_stack_contains_file(stack, filename)) {
		icon = kiba_stack_icon_new(stack);
		if (kiba_stack_icon_set_url (icon, filename)) {
			kiba_stack_add (stack, icon);
			kiba_stack_select_icon (stack, icon);
			file_added = TRUE;
		}
		else
			kiba_stack_icon_destroy (icon);
	}
	g_free (unescaped);
	
	/* select the icon-view item of the last added KibaStackIcon */
	if (KIBA_IS_ICON_VIEW_WIN(plugin->view_win) && plugin->view_stack == stack && icon) {
		KibaIconView *icon_view;
		
		icon_view = KIBA_ICON_VIEW(plugin->view_win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaStackIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
			if (check_icon == icon) {
				kiba_icon_view_select_child (
					KIBA_ICON_VIEW(stack->stack_plugin->view_win->icon_view),
					child);
				break;
			}
		}
	}
	
	return file_added;
}

static void unset_custom_icon (GtkWidget     *widget,
			       KibaStackIcon *icon) {
	g_free (icon->custom_icon);
	icon->custom_icon = NULL;
	kiba_stack_icon_update_surface (icon);
}

/**
activate a stack icon, launch the mime app for files and open the directory view for directory icons
**/
static void kiba_stack_activate_icon (KibaStackIcon *icon) {
	KibaAppChooser *chooser;
	gboolean custom_stack_icon;
	
	if (g_list_find (icon->stack->icons, icon))
		custom_stack_icon = TRUE;
	else
		custom_stack_icon = FALSE;
	
	/* the icon is a directory */
	if (icon->type == KIBA_STACK_ICON_TYPE_DIRECTORY) {
		if (icon->stack->stack_plugin->view_win &&
		    icon->stack != icon->stack->stack_plugin->view_stack) {
			kiba_stack_toggle_icon_view (icon->stack->stack_plugin->view_stack);
		}
		icon->stack->stack_plugin->view_stack = icon->stack;
		if (custom_stack_icon) {
			if (icon->stack->stack_plugin->view_icon) {
				unset_custom_icon (NULL, icon->stack->stack_plugin->view_icon);
				if (icon->stack->stack_plugin->view_win->win)
					g_signal_handlers_disconnect_by_func (
						G_OBJECT(icon->stack->stack_plugin->view_win->win),
						G_CALLBACK(unset_custom_icon),
						icon->stack->stack_plugin->view_icon);
			}
			icon->stack->stack_plugin->view_icon = icon;
		}
		create_icon_view_for_directory_icon (icon);
		if (icon->stack->stack_plugin->open_dirs && custom_stack_icon) {
			if (icon->custom_icon)
				g_free (icon->custom_icon);
			icon->custom_icon = g_strdup("folder-open");
			kiba_stack_icon_update_surface (icon);
			g_signal_connect (G_OBJECT(icon->stack->stack_plugin->view_win->win), "destroy",
				G_CALLBACK(unset_custom_icon), icon);
		}
	}
	/* a mime app was set, we can launch the file */
	else if (icon->exec) {
		kiba_launch_programm (icon->exec);
	}
	/* no mime app was associated previously, we gonna create a app-chooser */
	else {
		chooser = kiba_app_chooser_new(STACK_APPLICATION_DIR);
		gtk_widget_show (chooser->view_win->win);
		gdk_window_set_keep_above (chooser->view_win->win->window, TRUE);
		
		g_signal_connect (G_OBJECT(chooser), "app-activate",
			G_CALLBACK(kiba_stack_icon_app_selected), icon);
	}
}

/**
icon Info
**/

/**
Destroy the info win on each button press
**/
static void info_win_button_press_cb (GtkWidget      *widget,
				      GdkEventButton *button) {
	gtk_widget_destroy (widget);
}

/**
create a file-info popup (a "Info" menu item was pressed)
and place it at the current cursor position
**/
static void kiba_stack_icon_info_win (GtkWidget *widget,
				      gpointer   data) {
	static GtkWidget *info_win = NULL;
	static GtkWidget *mainbox = NULL;
	static GtkWidget *leftbox = NULL;
	static GtkWidget *rightbox = NULL;
	static GtkWidget *name_label = NULL;
	static GtkWidget *path_label = NULL;
	static GtkWidget *type_label = NULL;
	static GtkWidget *size_label = NULL;
	
	KibaStackIcon *icon = data;
	KibaScreen *screen;
	GtkWidget *label;
	GtkWidget *dumpbox;
	char *markup;
	const char *size_string;
	float file_size;
	float buf_size;
	struct stat status;
	
	screen = kiba_screen_get();
	
	if (!KIBA_IS_INFO_WIN(info_win)) {
		info_win = kiba_info_win_new ();
// 		KIBA_INFO_WIN(info_win)->gradient_start = icon->stack->stack_plugin->view_rgb1;
// 		KIBA_INFO_WIN(info_win)->gradient_mid   = icon->stack->stack_plugin->view_rgb2;
// 		KIBA_INFO_WIN(info_win)->gradient_end   = icon->stack->stack_plugin->view_rgb3;
// 		KIBA_INFO_WIN(info_win)->border_color   = icon->stack->stack_plugin->view_border_rgb;
		gtk_window_set_role (GTK_WINDOW(info_win), "kiba-stack");
		gtk_window_set_type_hint (GTK_WINDOW(info_win), GDK_WINDOW_TYPE_HINT_DOCK);
		gtk_window_set_decorated (GTK_WINDOW(info_win), FALSE);
		gtk_window_set_resizable (GTK_WINDOW(info_win), FALSE);
		
		mainbox = gtk_hbox_new(FALSE, 5);
		
		leftbox = gtk_vbox_new(TRUE, 5);
			label = gtk_label_new(_("Name"));
			markup = g_markup_printf_escaped (
				"<span weight=\"bold\" foreground=\"%s\">%s</span>",
				icon->stack->stack_plugin->text_color,
				GTK_LABEL (label)->text);
			gtk_label_set_markup (GTK_LABEL (label), markup);
			g_free (markup);
			gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
			gtk_widget_show (label);
			gtk_box_pack_start (GTK_BOX (leftbox), label, TRUE, TRUE, 0);
			
			label = gtk_label_new(_("Path"));
			markup = g_markup_printf_escaped (
				"<span weight=\"bold\" foreground=\"%s\">%s</span>",
				icon->stack->stack_plugin->text_color,
				GTK_LABEL (label)->text);
			gtk_label_set_markup (GTK_LABEL (label), markup);
			g_free (markup);
			gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
			gtk_widget_show (label);
			gtk_box_pack_start (GTK_BOX (leftbox), label, TRUE, TRUE, 0);
			
			label = gtk_label_new(_("Type"));
			markup = g_markup_printf_escaped (
				"<span weight=\"bold\" foreground=\"%s\">%s</span>",
				icon->stack->stack_plugin->text_color,
				GTK_LABEL (label)->text);
			gtk_label_set_markup (GTK_LABEL (label), markup);
			g_free (markup);
			gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
			gtk_widget_show (label);
			gtk_box_pack_start (GTK_BOX (leftbox), label, TRUE, TRUE, 0);
			
			label = gtk_label_new(_("Size"));
			markup = g_markup_printf_escaped (
				"<span weight=\"bold\" foreground=\"%s\">%s</span>",
				icon->stack->stack_plugin->text_color,
				GTK_LABEL (label)->text);
			gtk_label_set_markup (GTK_LABEL (label), markup);
			g_free (markup);
			gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
			gtk_widget_show (label);
			gtk_box_pack_start (GTK_BOX (leftbox), label, TRUE, TRUE, 0);
		gtk_box_pack_start (GTK_BOX (mainbox), leftbox, FALSE, FALSE, 0);
		gtk_widget_show (leftbox);
		
		rightbox = gtk_vbox_new(TRUE, 5);
			name_label = gtk_label_new(icon->name);
			gtk_misc_set_alignment (GTK_MISC(name_label), 0.0, 0.5);
			gtk_widget_show (name_label);
			gtk_box_pack_start (GTK_BOX (rightbox), name_label, TRUE, TRUE, 0);
			
			path_label = gtk_label_new(icon->url);
			gtk_misc_set_alignment (GTK_MISC(path_label), 0.0, 0.5);
			gtk_widget_show (path_label);
			gtk_box_pack_start (GTK_BOX (rightbox), path_label, TRUE, TRUE, 0);
			
			type_label = gtk_label_new(icon->mime_type);
			gtk_misc_set_alignment (GTK_MISC(type_label), 0.0, 0.5);
			gtk_widget_show (type_label);
			gtk_box_pack_start (GTK_BOX (rightbox), type_label, TRUE, TRUE, 0);
			
			size_label = gtk_label_new(_("Size"));
			gtk_misc_set_alignment (GTK_MISC(size_label), 0.0, 0.5);
			gtk_widget_show (size_label);
			gtk_box_pack_start (GTK_BOX (rightbox), size_label, TRUE, TRUE, 0);
		gtk_box_pack_start (GTK_BOX (mainbox), rightbox, FALSE, FALSE, 0);
		gtk_widget_show (rightbox);
		
		dumpbox = gtk_hbox_new(FALSE, 0);
		gtk_box_pack_start (GTK_BOX (dumpbox), mainbox, TRUE, TRUE, 5);
		gtk_widget_show (mainbox);
		GtkWidget *dumpbox2 = gtk_vbox_new(FALSE, 0);
		gtk_box_pack_start (GTK_BOX (dumpbox2), dumpbox, TRUE, TRUE, 5);
		gtk_widget_show (dumpbox);
		gtk_container_add (GTK_CONTAINER(info_win), dumpbox2);
		gtk_widget_show (dumpbox2);
		
		g_signal_connect (GTK_OBJECT(info_win), "button-press-event",
			G_CALLBACK(info_win_button_press_cb), NULL);
	}
	
	g_lstat (icon->url, &status);
	
	buf_size = (float) status.st_size;
	if (buf_size >= 1024.0) {
		buf_size /= 1024.0;
		if (buf_size >= 1024.0) {
			buf_size /= 1024.0;
			if (buf_size >= 1024.0) {
				buf_size /= 1024.0;
				size_string = _("GB");
			}
			else {
				size_string = _("MB");
			}
		}
		else {
			size_string = _("KB");
		}
	}
	else {
		size_string = _("Byte");
	}
	file_size = (float) ((int) (buf_size*10))/10.0;
	
	markup = g_markup_printf_escaped (
		"<span style=\"italic\" foreground=\"%s\">%s</span>",
		icon->stack->stack_plugin->text_color,
		icon->name);
	gtk_label_set_markup (GTK_LABEL (name_label), markup);
	g_free (markup);
	
	if (!g_utf8_validate (icon->url, -1, NULL)) {
		char *utf8_uri;
		
		utf8_uri = g_locale_to_utf8 (icon->url, -1, NULL, NULL, NULL);
		markup = g_markup_printf_escaped (
			"<span style=\"italic\" foreground=\"%s\">%s</span>",
			icon->stack->stack_plugin->text_color,
			utf8_uri);
		g_free (utf8_uri);
	}
	else {
		markup = g_markup_printf_escaped (
			"<span style=\"italic\" foreground=\"%s\">%s</span>",
			icon->stack->stack_plugin->text_color,
			icon->url);
	}
	gtk_label_set_markup (GTK_LABEL (path_label), markup);
	g_free (markup);
	
	markup = g_markup_printf_escaped (
		"<span style=\"italic\" foreground=\"%s\">%s</span>",
		icon->stack->stack_plugin->text_color,
		icon->mime_type);
	gtk_label_set_markup (GTK_LABEL (type_label), markup);
	g_free (markup);
	
	markup = g_markup_printf_escaped (
		"<span style=\"italic\" foreground=\"%s\">%g %s</span>",
		icon->stack->stack_plugin->text_color,
		file_size, size_string);
	gtk_label_set_markup (GTK_LABEL (size_label), markup);
	g_free (markup);
	
	gtk_window_present (GTK_WINDOW(info_win));
	gdk_window_set_keep_above (info_win->window, TRUE);
	
	gint x = 0, y = 0;
	gdk_window_get_pointer (gdk_get_default_root_window(), &x, &y, NULL);
	x -= info_win->allocation.width/2;
	y -= info_win->allocation.height/2;
	
	if (x < screen->geometry.x)
		x = screen->geometry.x;
	else if (x + info_win->allocation.width > screen->geometry.x + screen->geometry.width)
		x = screen->geometry.x + screen->geometry.width - info_win->allocation.width;
	
	if (y < screen->geometry.y)
		y = screen->geometry.y;
	else if (y + info_win->allocation.height > screen->geometry.y + screen->geometry.height)
		y = screen->geometry.y + screen->geometry.height - info_win->allocation.height;
	
	gtk_window_move (GTK_WINDOW(info_win),
		x,
		y);
}

/**
IconView
**/

/**
Place the icon view on the object (below for a top-dock, above for a bottom-dock ....)
**/
static void kiba_stack_place_icon_view (KibaStackPlugin *plugin,
					GtkWidget       *widget) {
	KibaWin *win;
	KibaObject *object;
	KibaScreen *screen;
	gint w, h, x, y;
	const KibaLayoutData *d;
	
	screen = kiba_screen_get();
	win = kiba_win_get_default();
	object = KIBA_OBJECT(widget);
	
	if (!KIBA_IS_WIN(win))
		return;
	if (!KIBA_IS_ICON_VIEW_WIN(plugin->view_win))
		return;
	if (!GTK_WIDGET_REALIZED(plugin->view_win->win))
		return;
	
	d = &kiba_layout_data[KIBA_DOCK(KIBA_WIN_CHILD(win))->d.position];
	
	struct {
		gint  *x, *y;
		gint  *w, *h;
		gint  alignment;
	} magic[2] = {
		{ &x, &y , &w, &h, d->valignment},
		{ &y, &x , &h, &w, d->halignment}
	};
	w = plugin->view_win->win->allocation.width;
	h = plugin->view_win->win->allocation.height;
	x = widget->allocation.x + GTK_WIDGET(win)->allocation.x;
	y = widget->allocation.y + GTK_WIDGET(win)->allocation.y;
	
	*magic[d->orientation].x -= (*magic[d->orientation].w - widget->allocation.width)/2;
	*magic[d->orientation].y -=
		(magic[d->orientation].alignment == 0 ? - widget->allocation.height : *magic[d->orientation].h);
	
	if (x < screen->geometry.x)
		x = screen->geometry.x;
	else if (x + plugin->view_win->win->allocation.width > screen->geometry.x + screen->geometry.width)
		x = screen->geometry.x + screen->geometry.width - plugin->view_win->win->allocation.width;
	
	if (y < screen->geometry.y)
		y = screen->geometry.y;
	else if (y + plugin->view_win->win->allocation.height >
		 screen->geometry.y + screen->geometry.height)
		y = screen->geometry.y + screen->geometry.height -
			plugin->view_win->win->allocation.height;
	
	gdk_window_move (plugin->view_win->win->window, x, y);
}

static void kiba_stack_icon_view_size_allocate_cb (GtkWidget     *widget,
						   GtkAllocation *allocation,
						   gpointer       data) {
	KibaStack *stack = data;
	
	kiba_stack_place_icon_view (stack->stack_plugin, stack->widget);
}

static GtkWidget* kiba_stack_create_context_menu_for_icon (KibaStackIcon *icon) {
	GtkWidget *menu = NULL;
	GtkWidget *item;
	GtkWidget *image;
	
	menu = gtk_menu_new ();
	
	item = gtk_image_menu_item_new_with_label (_("Info"));
	image = gtk_image_new_from_icon_name (GTK_STOCK_INFO, GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	g_signal_connect (G_OBJECT (item), "activate",
		G_CALLBACK (kiba_stack_icon_info_win), icon);
	gtk_widget_show (item);
	
	if (icon->type != KIBA_STACK_ICON_TYPE_DIRECTORY &&
	    icon->type != KIBA_STACK_ICON_TYPE_DESKTOP) {
		item = gtk_image_menu_item_new_with_label (_("Set MIME Application"));
// 		image = gtk_image_new_from_icon_name (GTK_STOCK_REFRESH, GTK_ICON_SIZE_MENU);
// 		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		g_signal_connect (G_OBJECT (item), "activate",
			G_CALLBACK (kiba_stack_set_mime_app_cb), icon);
		gtk_widget_show (item);
	}
	
	if (g_list_find(icon->stack->icons, icon)) {
		if (KIBA_IS_STACK(icon->stack))
			kiba_menu_add_separator (GTK_WIDGET (menu));
		
		item = gtk_image_menu_item_new_with_label (_("Remove Stack Item"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		g_signal_connect (G_OBJECT (item),
			"activate", G_CALLBACK (kiba_stack_remove_icon_cb), icon);
		gtk_widget_show (item);
	}
	
	return menu;
}

static gboolean kiba_stack_button_press_cb (GtkWidget      *widget,
					    GdkEventButton *event,
					    gpointer        data) {
	KibaStack *stack = data;
	
	if (event->button == 3) {
		GtkWidget *menu;
		GtkWidget *item;
		GtkWidget *image;
		
		menu = kiba_stack_create_context_menu_for_icon (stack->icons->data);
		if (!menu)
			return FALSE;
		
		item = gtk_image_menu_item_new_with_label (_("Remove Stack"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		g_signal_connect (G_OBJECT (item), "activate",
			G_CALLBACK (kiba_stack_remove_cb), stack);
		gtk_widget_show (item);
	
		gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
			event->button, gtk_get_current_event_time());
		
		return TRUE;
	}
//TODO item for directory icons to show the folder in nautilus
	
	return FALSE;
}

static gboolean kiba_stack_view_button_press_cb (GtkWidget      *widget,
					         GdkEventButton *event,
					         gpointer        data) {
	KibaStackIcon *icon = data;
	
	if (event->button == 3) {
		GtkWidget *menu;
		
		menu = kiba_stack_create_context_menu_for_icon (icon);
		if (!menu)
			return FALSE;
		
		gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
			event->button, gtk_get_current_event_time());
	}
	
	return FALSE;
}

/**
a drag move was completed,
we gonna remove the KibaStackIcon and its icon-view icon (if there is one)
**/
static void kiba_icon_view_drag_delete_cb (GtkWidget      *widget,
					   GdkDragContext *drag_context,
					   gpointer        user_data) {
	KibaStackIcon *icon = user_data;
	KibaIconView *icon_view;
	GList *l;
	
	icon_view = KIBA_ICON_VIEW(icon->stack->stack_plugin->view_win->icon_view);
	
	for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		KibaStackIcon *check_icon;
		
		check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
		if (check_icon == icon) {
			if (g_list_length(l) == 1) {
				g_signal_handlers_disconnect_by_func (
				     G_OBJECT(icon->stack->widget),
				     G_CALLBACK(kiba_stack_icon_view_size_allocate_cb),
				     icon->stack);
				g_signal_handlers_disconnect_by_func (
				     G_OBJECT(icon->stack->stack_plugin->view_win->win),
				     G_CALLBACK(kiba_stack_icon_view_size_allocate_cb),
				     icon->stack);
			}
			
			kiba_icon_view_win_remove (icon->stack->stack_plugin->view_win, child);
			break;
		}
	}
	kiba_stack_remove_icon_cb (NULL, icon);
}

/**
append the icon uri to the drag path
**/
static void kiba_icon_view_drag_data_get_cb (GtkWidget        *widget,
					     GdkDragContext   *drag_context,
					     GtkSelectionData *data,
					     guint             info,
					     guint             time,
					     gpointer          user_data) {
	KibaStackIcon *icon = user_data;
	GString *result;
	char *string;
	
	result = g_string_new (NULL);
	
	string = gnome_vfs_make_uri_from_shell_arg (icon->url);
	g_string_append (result, string);
	g_free (string);
	g_string_append (result, "\r\n");
	
	gtk_selection_data_set (data,
				data->target,
				8, result->str, result->len);
}

/**
setting the drag action the view wanna take
**/
static gboolean kiba_icon_view_drag_motion (GtkWidget      *widget,
					    GdkDragContext *dc,
					    gint            x,
					    gint            y,
					    guint           t,
					    gpointer        data) {
	KibaStack *stack = data;
	GdkDragAction suggested_action;
	GtkWidget *src_widget;
	
	/* Get source widget */
	src_widget = gtk_drag_get_source_widget(dc);
	
	if (src_widget == widget) /* source == target, skip this one */
		suggested_action = 0;
	/* expect moving for directories */
	else if (stack->stack_plugin->view_type == KIBA_STACK_TYPE_DIRECTORY)
		suggested_action = GDK_ACTION_MOVE;
	else
		suggested_action = GDK_ACTION_COPY;
	
	/* set the drag status */
	if(dc->actions == GDK_ACTION_MOVE) {
		gdk_drag_status(dc, GDK_ACTION_MOVE, t);
	}
	else if(dc->actions == GDK_ACTION_COPY) {
		gdk_drag_status(dc, GDK_ACTION_COPY, t);
	}
	else if(dc->actions == GDK_ACTION_LINK) {
		gdk_drag_status(dc, GDK_ACTION_LINK, t);
	}
	else if(dc->actions & suggested_action) {
		gdk_drag_status(dc, suggested_action, t);
	}
	else {
		gdk_drag_status(dc, 0, t);
	}
	
	return FALSE;
}

/**
set markup for the icon-view item titles
**/
static void image_box_label_set_markup (KibaStackPlugin *plugin,
					GtkWidget       *box,
					GtkWidget       *label) {
	KibaIconView *icon_view;
	char *markup;
	
	icon_view = KIBA_ICON_VIEW(plugin->view_win->icon_view);
	
	if (box == icon_view->selected_child) {
		markup = g_markup_printf_escaped (
			"<span size=\"small\" style=\"%s\" weight=\"bold\" foreground=\"%s\">%s</span>",
			(plugin->text_italic == TRUE ? "italic" : "normal"),
			 plugin->text_color,
			GTK_LABEL (label)->text);
	}
	else {
		markup = g_markup_printf_escaped (
			"<span size=\"small\" style=\"%s\" foreground=\"%s\">%s</span>",
			(plugin->text_italic == TRUE ? "italic" : "normal"),
			 plugin->text_color,
			 GTK_LABEL (label)->text);
	}
	gtk_label_set_markup (GTK_LABEL (label), markup);
	g_free (markup);
}

static gboolean kiba_icon_view_dnd_expose_transparent (GtkWidget      *widget,
						       GdkEventExpose *expose) {
	cairo_t *cr = NULL;
	cairo_pattern_t *pattern = NULL;
	GtkWidget *child = NULL;
	
	child = gtk_bin_get_child (GTK_BIN (widget));
	
	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;
	
	cr = gdk_cairo_create (widget->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_paint (cr);
	cairo_destroy (cr);
	
	if (child)
		gtk_container_propagate_expose (GTK_CONTAINER (widget), child,  expose);
	
	return FALSE;
}

static void kiba_icon_view_dnd_image_screen_changed (GtkWidget *widget,
						     GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	screen = gtk_widget_get_screen (widget);
	
	colormap = gdk_screen_get_rgba_colormap (screen);
	visual   = gdk_screen_get_rgba_visual (screen);
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual   = gdk_screen_get_rgb_visual (screen);
	}
	
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

static void kiba_icon_view_dnd_image_drag_end_cb (GtkWidget      *widget,
						  GdkDragContext *drag_context,
						  gpointer        user_data) {
	GtkWidget *image = user_data;
	
	gtk_widget_destroy (image);
}

static void unset_dnd_icon (GtkWidget      *widget,
			    GdkDragContext *drag_context) {
	stack_dnd_icon = NULL;
}

static void kiba_icon_view_drag_begin_cb (GtkWidget      *widget,
					  GdkDragContext *drag_context,
					  gpointer        data) {
	KibaStackIcon *icon = data;
	GtkWidget *image;
	GtkWidget *win;
	GtkWidgetClass *widget_class;
	
	win = gtk_window_new (GTK_WINDOW_POPUP);
	widget_class = GTK_WIDGET_GET_CLASS(win);
	widget_class->expose_event = kiba_icon_view_dnd_expose_transparent;
	widget_class->screen_changed = kiba_icon_view_dnd_image_screen_changed;
	kiba_icon_view_dnd_image_screen_changed (win, NULL);
	gtk_window_set_type_hint (GTK_WINDOW(win), GDK_WINDOW_TYPE_HINT_DND);
	
	image = kiba_image_new ();
	kiba_image_set_surface (KIBA_IMAGE(image), icon->surface->surface,
			icon->surface->width,
			icon->surface->height);
	kiba_image_set_size (KIBA_IMAGE(image),
			icon->surface->width,
			icon->surface->height);
	gtk_widget_show (image);
	
	gtk_container_add (GTK_CONTAINER(win), image);
	
	gtk_drag_set_icon_widget (drag_context, win, 0, 0);
	
	stack_dnd_icon = icon;
	
	g_signal_connect (G_OBJECT(widget), "drag-end",
		G_CALLBACK(unset_dnd_icon), NULL);
}

/**
create a icon-view item for a KibaStackIcon,
using KibaImage for the icon and GtkLabel for the text
**/
static GtkWidget* create_image_box_for_icon (KibaStackIcon *icon,
					     gint           size) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	GtkWidget *image;
	GtkWidget *box;
	GtkWidget *label;
	char *markup;
	
	box = gtk_vbox_new(FALSE, 0);
	
	GtkWidget *image_table;
	image_table = gtk_table_new (1, 3, FALSE);
	
	image = kiba_image_new ();
	kiba_image_set_surface (KIBA_IMAGE(image), icon->surface->surface,
			icon->surface->width, icon->surface->height);
	kiba_image_set_size (KIBA_IMAGE(image), size, size);
	gtk_widget_show (image);
	gtk_table_attach (GTK_TABLE(image_table), image,
				1, 2,
				0, 1,
				GTK_EXPAND | GTK_SHRINK, GTK_EXPAND | GTK_SHRINK,
				0, 0);
	gtk_box_pack_start (GTK_BOX (box), image_table, TRUE, TRUE, 0);
	gtk_widget_show (image_table);
	
	label = gtk_label_new(icon->name);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	gtk_label_set_ellipsize (GTK_LABEL(label), PANGO_ELLIPSIZE_END);
	image_box_label_set_markup (icon->stack->stack_plugin, box, label);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
	
	gtk_drag_source_set (image,
		GDK_BUTTON1_MASK | GDK_BUTTON2_MASK,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_MOVE | GDK_ACTION_COPY);
	
	g_signal_connect (GTK_OBJECT(image), "drag-data-get",
		G_CALLBACK(kiba_icon_view_drag_data_get_cb), icon);
	g_signal_connect (GTK_OBJECT(image), "drag-data-delete",
		G_CALLBACK(kiba_icon_view_drag_delete_cb), icon);
	g_signal_connect (GTK_OBJECT(image), "drag-begin",
		G_CALLBACK(kiba_icon_view_drag_begin_cb), icon);
	g_signal_connect (GTK_OBJECT(image), "button-press-event",
		G_CALLBACK(kiba_stack_view_button_press_cb), icon);
	
	g_object_set_data (G_OBJECT(box), "stack-icon", icon);
	
	gtk_widget_set_usize (box, size + 30, -1);
	
	return box;
}

typedef struct {
	KibaStack           *stack;
	GnomeVFSAsyncHandle *handle_return;
	GdkDragContext      *context;
	GtkWidget           *box;
	char                *url;
}KibaStackXferData;

static gint  kiba_stack_xfer_progress_callback (GnomeVFSAsyncHandle      *handle,
						GnomeVFSXferProgressInfo *info,
						gpointer                  user_data) {
	KibaStackXferData *data = user_data;
	
	if (info->phase >= GNOME_VFS_XFER_PHASE_INITIAL && !data->box) {
		GtkWidget *image;
		GtkWidget *label;
		
		data->box = gtk_vbox_new (FALSE, 0);
		
		image = gtk_image_new_from_stock (GTK_STOCK_REFRESH, GTK_ICON_SIZE_DIALOG);
		gtk_widget_set_usize (image,
			data->stack->stack_plugin->view_icon_size,
			data->stack->stack_plugin->view_icon_size);
		gtk_widget_show (image);
		gtk_box_pack_start (GTK_BOX (data->box), image, FALSE, FALSE, 0);
		
		label = gtk_label_new (g_strrstr(data->url, "/")+1);
		gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
		gtk_label_set_ellipsize (GTK_LABEL(label), PANGO_ELLIPSIZE_END);
		image_box_label_set_markup (data->stack->stack_plugin, data->box, label);
		gtk_widget_show (label);
		gtk_box_pack_start (GTK_BOX (data->box), label, FALSE, FALSE, 0);
		
		kiba_icon_view_win_add (data->stack->stack_plugin->view_win, data->box);
	}
	
	if (info->phase == GNOME_VFS_XFER_PHASE_COMPLETED) {
		gtk_widget_destroy (data->box);
		create_icon_for_uri (data->stack, data->url);
		g_free (data->url);
		g_free (data);
	}
	
	return 1;
}

/**
a directory icon-view received a drag,
we gonna copy/link or move the file to the directory
**/
static gboolean kiba_stack_extern_drag_receive_cb (KibaStack      *stack,
						   GdkDragContext *drag_context,
						   const char     *filepath) {
	KibaIconView *icon_view;
	GnomeVFSXferOptions xfer_options;
	GList *l;
	char *new_path;
	char *file_name;
	gboolean async_xfer_callback;
	
	icon_view = KIBA_ICON_VIEW(stack->stack_plugin->view_win->icon_view);
	
	if (!g_file_test(filepath, G_FILE_TEST_IS_DIR))
		file_name = g_strrstr(filepath, "/")+1;
	else
		file_name = "";
	new_path = g_strdup_printf ("%s/%s", stack->stack_plugin->view_path, file_name);
	
	for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		KibaStackIcon *icon = g_object_get_data (G_OBJECT(child), "stack-icon");
		
		if (g_strrstr (icon->url, new_path)) {
			g_free (new_path);
			return 0;
		}
	}
	
	async_xfer_callback = FALSE;
	switch (drag_context->action) {
		case GDK_ACTION_MOVE: {
			async_xfer_callback = TRUE;
			xfer_options = GNOME_VFS_XFER_REMOVESOURCE;
			break;
		}
		case GDK_ACTION_COPY: {
			xfer_options = GNOME_VFS_XFER_DEFAULT;
			async_xfer_callback = TRUE;
			break;
		}
		case GDK_ACTION_LINK: {
			link (filepath, new_path);
			create_icon_for_uri (stack, new_path);
			break;
		}
	}
	
	if (async_xfer_callback) {
		KibaStackXferData *xfer_data;
		GnomeVFSURI *source_uri;
		GnomeVFSURI *target_uri;
		GList *src_list;
		GList *dest_list;
		char *escaped_uri;
		
		escaped_uri = gnome_vfs_make_uri_from_input (filepath);
		source_uri = gnome_vfs_uri_new (escaped_uri);
		g_free (escaped_uri);
		escaped_uri = gnome_vfs_make_uri_from_input (new_path);
		target_uri = gnome_vfs_uri_new (escaped_uri);
		g_free (escaped_uri);
		
		src_list  = g_list_append (NULL, source_uri);
		dest_list = g_list_append (NULL, target_uri);
		
		xfer_data = g_new0 (KibaStackXferData, 1);
		xfer_data->stack = stack;
		xfer_data->context = drag_context;
		xfer_data->box = NULL;
		xfer_data->url = g_strdup (new_path);
		xfer_data->handle_return = NULL;
		
		gnome_vfs_async_xfer (&xfer_data->handle_return,
				      src_list,
				      dest_list,
				      xfer_options,
				      GNOME_VFS_XFER_ERROR_MODE_ABORT,
				      GNOME_VFS_XFER_OVERWRITE_MODE_ABORT,
				      GNOME_VFS_PRIORITY_DEFAULT,
				      kiba_stack_xfer_progress_callback,
				      xfer_data,
				      NULL,
				      NULL);
	}
	
	g_free (new_path);
	
	return 1;
}

/**
a icon-view received a drag
**/
static void kiba_icon_view_drag_received_cb (GtkWidget        *widget,
					     GdkDragContext   *drag_context,
					     gint              x,
					     gint              y,
					     GtkSelectionData *selection_data,
					     guint             info,
					     guint             time,
					     gpointer          user_data) {
	static const char *file_prefix = "file://";
	
	KibaStack *stack = user_data;
	
	if (stack->stack_plugin->view_type == KIBA_STACK_TYPE_CUSTOM) {
		if (kiba_stack_drag_receive_cb (stack, drag_context, selection_data))
			gtk_drag_finish (drag_context,
				TRUE,
				FALSE, time);
		else {
			gdk_drag_status (drag_context, 0, time);
		}
	}
	/* the icon-view is a directory */
	else {
		char *filepath;
		char *string;
		char *unescaped;
		gint new_icons;
		
		char data[strlen((char*) selection_data->data)];
		sprintf(data, "%s", (char*) selection_data->data);
		g_strstrip (data);
		
		new_icons = 0;
		while ((string = g_strrstr ((char*) data, "\n")) != NULL) {
			g_strstrip (string);
			
			if (g_str_has_prefix(string, file_prefix)) {
				unescaped = gnome_vfs_unescape_string (string + strlen(file_prefix), NULL);
			}
			else {
				unescaped = gnome_vfs_unescape_string (string, NULL);
			}
			filepath = unescaped;
			
			if (g_file_test(filepath, G_FILE_TEST_EXISTS)) {
				new_icons += kiba_stack_extern_drag_receive_cb (stack,
						drag_context, filepath);
			}
			
			string[0] = '\0';
		}
		g_strstrip (data);
		if (g_str_has_prefix(data, file_prefix)) {
			unescaped = gnome_vfs_unescape_string (data + strlen(file_prefix), NULL);
		}
		else {
			unescaped = gnome_vfs_unescape_string (data, NULL);
		}
		filepath = unescaped;
		
		if (g_file_test(filepath, G_FILE_TEST_EXISTS)) {
			new_icons += kiba_stack_extern_drag_receive_cb (stack, drag_context, filepath);
		}
		g_free (unescaped);
		
		if (new_icons > 0)
			gtk_drag_finish (drag_context,
				TRUE,
				FALSE, /* currently we remove the file on our own,
					  but we should involve the drag-data-delete event */
				time);
		else
			gdk_drag_status (drag_context, 0, time);
	}
}

static void kiba_icon_view_child_activated_cb (KibaIconView *icon_view,
					       GtkWidget    *child,
					       gpointer      user_data) {
	kiba_stack_activate_icon (g_object_get_data (G_OBJECT(child), "stack-icon"));
}

static void kiba_icon_view_child_selected_cb (KibaIconView *icon_view,
					      GtkWidget    *child,
					      gpointer      user_data) {
	KibaStack *stack = user_data;
	KibaStackIcon *selected;
	GList *l;
	
	selected = g_object_get_data (G_OBJECT(child), "stack-icon");
	
	for (l = gtk_container_get_children(GTK_CONTAINER(child)); l != NULL; l = l->next) {
		GtkWidget *widget = l->data;
		
		if (!GTK_IS_LABEL(widget))
			continue;
		
		image_box_label_set_markup(stack->stack_plugin, child, widget);
	}
	
	if (stack->stack_plugin->view_type == KIBA_STACK_TYPE_CUSTOM)
		kiba_stack_select_icon (stack, selected);
}

static void kiba_icon_view_child_unselected_cb (KibaIconView *icon_view,
					        GtkWidget    *child,
					        gpointer      user_data) {
	KibaStack *stack = user_data;
	GList *l;
	
	for (l = gtk_container_get_children(GTK_CONTAINER(child)); l != NULL; l = l->next) {
		GtkWidget *widget = l->data;
		
		if (!GTK_IS_LABEL(widget))
			continue;
		
		image_box_label_set_markup(stack->stack_plugin, child, widget);
	}
}

/**
clean up invalid pointers after the icon view win was destroyed
**/
static void kiba_stack_win_destroyed_cb (GtkWidget *widget,
					 gpointer   data) {
	KibaStack *stack = data;
	
	g_signal_handlers_disconnect_by_func (
	     G_OBJECT(stack->widget),
	     G_CALLBACK(kiba_stack_icon_view_size_allocate_cb),
	     stack);
	
	stack->stack_plugin->view_icon = NULL;
	stack->stack_plugin->view_win = NULL;
	stack->stack_plugin->view_type = KIBA_STACK_TYPE_NONE;
	stack->stack_plugin->view_stack = NULL;
	if (stack->stack_plugin->view_path)
		g_free (stack->stack_plugin->view_path);
	stack->stack_plugin->view_path = NULL;
}

/**
create a icon-view dialog window
**/
static KibaIconViewWin* kiba_stack_plugin_create_icon_view (KibaStack *stack) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaIconViewWin *view_win;
	KibaInfoWin *info_win;
	KibaIconView *icon_view;
	
	view_win = kiba_icon_view_win_new(KIBA_ICON_VIEW_BOX_POSITION_RIGHT,
			stack->stack_plugin->view_columns,
			stack->stack_plugin->view_rows, TRUE);
	gtk_window_set_type_hint (GTK_WINDOW(view_win->win), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_window_stick (GTK_WINDOW(view_win->win));
	
	info_win = KIBA_INFO_WIN(view_win->win);
	icon_view = KIBA_ICON_VIEW(view_win->icon_view);
	icon_view->slide_step = stack->stack_plugin->slide_step;
// 	info_win->gradient_start = stack->stack_plugin->view_rgb1;
// 	info_win->gradient_mid   = stack->stack_plugin->view_rgb2;
// 	info_win->gradient_end   = stack->stack_plugin->view_rgb3;
// 	info_win->border_color   = stack->stack_plugin->view_border_rgb;
	gtk_window_set_role (GTK_WINDOW(view_win->win), "kiba-stack");
	
	gtk_drag_dest_set (view_win->icon_view,
		GTK_DEST_DEFAULT_ALL,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_MOVE);
	gtk_drag_dest_add_uri_targets (view_win->icon_view);
	
	/* adding signal handlers */
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "child-activated",
		G_CALLBACK(kiba_icon_view_child_activated_cb), NULL);
	g_signal_connect(G_OBJECT (stack->widget), "size-allocate",
		G_CALLBACK(kiba_stack_icon_view_size_allocate_cb), stack);
	g_signal_connect_after(G_OBJECT (view_win->icon_view), "size-allocate",
		G_CALLBACK(kiba_stack_icon_view_size_allocate_cb), stack);
	g_signal_connect (G_OBJECT(view_win->win), "destroy",
		G_CALLBACK(kiba_stack_win_destroyed_cb), stack);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "drag-data-received",
		G_CALLBACK(kiba_icon_view_drag_received_cb), stack);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "drag-motion",
		G_CALLBACK(kiba_icon_view_drag_motion), stack);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "child-selected",
		G_CALLBACK(kiba_icon_view_child_selected_cb), stack);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "child-unselected",
		G_CALLBACK(kiba_icon_view_child_unselected_cb), stack);
	
	stack->stack_plugin->view_win = view_win;
	
	return view_win;
}

/**
create a custom icon view win, containing all files/dirs in the stack
**/
static KibaIconViewWin* kiba_stack_create_icon_view_win (KibaStack *stack) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaIconViewWin *view_win;
	GtkWidget *box;
	GList *l;
	GList *sorted;
	
	sorted = g_list_sort (g_list_copy(stack->icons), kiba_stack_sort_icons_by_name);
	view_win = kiba_stack_plugin_create_icon_view (stack);
	for (l = sorted; l != NULL; l = l->next) {
		KibaStackIcon *icon = l->data;
		
		box = create_image_box_for_icon (icon, stack->stack_plugin->view_icon_size);
		kiba_icon_view_win_add (view_win, box);
		if (icon == stack->icons->data) {
			// TODO: the currently selected icon is maybe unshown
			// kiba_icon_view_win_show_child (view_win->icon_view, box);
			kiba_icon_view_select_child (
				KIBA_ICON_VIEW(view_win->icon_view),
				box);
		}
	}
	g_list_free (sorted);
	
	stack->stack_plugin->view_type = KIBA_STACK_TYPE_CUSTOM;
	stack->stack_plugin->view_stack = stack;
	if (stack->stack_plugin->view_path)
		g_free (stack->stack_plugin->view_path);
	stack->stack_plugin->view_path = g_strdup (stack->uri);
	
	return view_win;
}

void kiba_stack_toggle_icon_view (KibaStack *stack) {
	GList *l;
	
	if (KIBA_IS_ICON_VIEW_WIN(stack->stack_plugin->view_win)) {
		gboolean toggling;
		
		toggling = (stack->stack_plugin->view_stack == stack);
		
		gtk_widget_destroy (stack->stack_plugin->view_win->win);
		stack->stack_plugin->view_type = KIBA_STACK_TYPE_NONE;
		if (stack->stack_plugin->view_path)
			g_free (stack->stack_plugin->view_path);
		stack->stack_plugin->view_path = NULL;
		
		if (toggling) {
			stack->stack_plugin->view_stack = NULL;
			return;
		}
	}
	
	stack->stack_plugin->view_win = kiba_stack_create_icon_view_win (stack);
	gtk_widget_show (stack->stack_plugin->view_win->win);
	gdk_window_set_keep_above (stack->stack_plugin->view_win->win->window, TRUE);
	kiba_stack_place_icon_view (stack->stack_plugin, stack->widget);
}

/**
Directory Icon View
**/

static void kiba_stack_dir_view_destroyed_cb (GtkWidget *widget,
					      gpointer   data) {
	KibaStackIcon *icon = data;
	
	kiba_stack_icon_destroy (icon);
}

/**
create a KibaStackIcon for the given uri and put it in the view win
**/
static GtkWidget* create_icon_for_uri (KibaStack   *stack,
				       const char  *uri) {
	GtkWidget *ret;
	KibaStackIcon *icon;
	
	icon = kiba_stack_icon_new (stack);
	if (!kiba_stack_icon_set_url (icon, uri)) {
		kiba_stack_icon_destroy (icon);
		return NULL;
	}
	icon->surface->width   = stack->stack_plugin->view_icon_size;
	icon->surface->height  = stack->stack_plugin->view_icon_size;
	kiba_stack_icon_update_surface (icon);
	
	ret = create_image_box_for_icon (icon, stack->stack_plugin->view_icon_size);
	kiba_icon_view_win_add (stack->stack_plugin->view_win, ret);
	
	g_signal_connect (G_OBJECT(ret), "destroy",
		G_CALLBACK(kiba_stack_dir_view_destroyed_cb), icon);
	
	return ret;
}

/**
create a directory icon view, containing all files of the directory and a "..." item
**/
static gboolean create_icon_view_for_directory_icon (KibaStackIcon *directory_icon) {
	GtkWidget *box = NULL;
	GError *error = NULL;
	char *file_name;
	char *parent_dir;
	char *parent_nedle;
	GDir *dir;
	GList *l;
	GList *pathes = NULL;
	
	dir = g_dir_open (directory_icon->url, 0, &error);
	if (error != NULL) {
		g_error_free (error);
		return FALSE;
	}
	
	if (!KIBA_IS_ICON_VIEW_WIN(directory_icon->stack->stack_plugin->view_win)) {
		kiba_stack_plugin_create_icon_view (directory_icon->stack);
	}
	else {
		KibaIconView *icon_view;
		icon_view = KIBA_ICON_VIEW(directory_icon->stack->stack_plugin->view_win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(icon_view));
		     l != NULL; l = gtk_container_get_children(GTK_CONTAINER(icon_view))) {
			kiba_icon_view_win_remove (directory_icon->stack->stack_plugin->view_win,
				GTK_WIDGET(l->data));
		}
	}
	directory_icon->stack->stack_plugin->view_type = KIBA_STACK_TYPE_DIRECTORY;
	directory_icon->stack->stack_plugin->view_stack = directory_icon->stack;
	if (directory_icon->stack->stack_plugin->view_path)
		g_free (directory_icon->stack->stack_plugin->view_path);
	directory_icon->stack->stack_plugin->view_path = g_strdup (directory_icon->url);
	
	parent_nedle = g_strrstr (directory_icon->url, "/");
	if (parent_nedle) {
		if (parent_nedle != directory_icon->url) {
			parent_dir = g_strndup (directory_icon->url,
					strlen(directory_icon->url) - strlen(parent_nedle));
			box = create_icon_for_uri (directory_icon->stack, parent_dir);
			g_free (parent_dir);
		}
		else if (strcmp (parent_nedle, "/") != 0)
			box = create_icon_for_uri (directory_icon->stack, "/");
		
		if (box) {
			for (l = gtk_container_get_children(GTK_CONTAINER(box)); l != NULL; l = l->next) {
				GtkWidget *child = l->data;
				char *markup;
				
				if (!GTK_IS_LABEL(child))
					continue;
				
				gtk_label_set_text (GTK_LABEL(child), "...");
				markup = g_markup_printf_escaped (
					"<span size=\"small\" foreground=\"%s\">%s</span>",
					directory_icon->stack->stack_plugin->text_color,
					GTK_LABEL (child)->text);
				gtk_label_set_markup (GTK_LABEL (child), markup);
				g_free (markup);
			}
		}
	}
	
	while ((file_name = (char*) g_dir_read_name (dir)) != NULL) {
		char *uri;
		
		/* skipping hidden files */
		if (file_name[0] == '.')
			continue;
		
		if (strcmp (directory_icon->url, "/") == 0)
			uri = g_strdup_printf ("/%s", file_name);
		else
			uri = g_strdup_printf ("%s/%s", directory_icon->url, file_name);
		
		pathes = g_list_append (pathes, uri);
	}
	g_dir_close (dir);
	
	pathes = g_list_sort (pathes, kiba_utils_sort_string_list);
	
	for (l = pathes; l != NULL; l = pathes) {
		char *path = l->data;
		
		box = create_icon_for_uri (directory_icon->stack, path);
		g_free (path);
		pathes = g_list_remove (pathes, l->data);
	}
	g_list_free (pathes);
	
	gtk_window_present (GTK_WINDOW(directory_icon->stack->stack_plugin->view_win->win));
	gdk_window_set_keep_above (directory_icon->stack->stack_plugin->view_win->win->window, TRUE);
	kiba_stack_place_icon_view (directory_icon->stack->stack_plugin, directory_icon->stack->widget);
	
	return TRUE;
}

/**
KibaObject
**/

static void kiba_object_drag_received_cb (GtkWidget        *widget,
					  GdkDragContext   *drag_context,
					  gint              x,
					  gint              y,
					  GtkSelectionData *selection_data,
					  guint             info,
					  guint             time,
					  gpointer          user_data) {
	KibaStack *stack = user_data;
	
	gdk_drag_status (drag_context, GDK_ACTION_COPY, time);
	kiba_stack_drag_receive_cb (stack, drag_context, selection_data);
	gtk_drag_finish (drag_context,
		TRUE,
		FALSE, time);
	
	if (KIBA_IS_STACK_ICON(stack_dnd_icon) &&
	    stack->stack_plugin->view_type != KIBA_STACK_TYPE_DIRECTORY) {
		kiba_icon_view_drag_delete_cb (NULL, NULL, stack_dnd_icon);
	}
}

static gboolean kiba_stack_button_release_cb (GtkWidget      *widget,
					      GdkEventButton *event,
					      gpointer        data) {
	KibaStack *stack = data;
	KibaStackIcon *icon;
	
	if (event->button == 3)
		return TRUE;
	
	icon = stack->icons->data;
	
	if (event->button == 2) {
		if (stack->stack_plugin->view_icon == icon)
			gtk_widget_destroy (stack->stack_plugin->view_win->win);
		else
			kiba_stack_activate_icon (icon);
		
		return TRUE;
	}
	
	kiba_stack_toggle_icon_view (stack);
	
	return TRUE;
}

static void kiba_stack_scroll_cb (GtkWidget      *widget,
				  GdkEventScroll *event,
				  gpointer        data) {
	KibaStack *stack = data;
	GList *l;
	
	if (stack->count <= 1)
		return;
	
	kiba_object_switch_surface (KIBA_OBJECT(widget));
	
	kiba_stack_icon_set_state (KIBA_STACK_ICON(stack->icons->data),
			KIBA_STACK_STATE_INACTIVE);
	kiba_utils_glist_scroll (&stack->icons, event->direction);
	kiba_stack_icon_set_state (KIBA_STACK_ICON(stack->icons->data),
		KIBA_STACK_STATE_ACTIVE);
	
	if (stack->stack_plugin->view_type == KIBA_STACK_TYPE_CUSTOM) {
		kiba_icon_view_scroll_cb (stack->stack_plugin->view_win->icon_view,
			event, stack->stack_plugin->view_win);
	}
}

/**
Thumbnailing

Note: based on nautilus thumbnailing
**/

#include <pthread.h>
/* Cool-off period between last file modification time and thumbnail creation */
#define THUMBNAIL_CREATION_DELAY_SECS 3
/* Should never be a reasonable actual mtime */
#define INVALID_MTIME 0

/* structure used for making thumbnails, associating a uri with where the thumbnail is to be stored */
typedef struct {
	KibaStackIcon         *icon;
	GnomeThumbnailFactory *thumb_factory;
	char                  *image_uri;
	char                  *mime_type;
	time_t                 original_file_mtime;
} KibaThumbnailInfo;

/* Our mutex used when accessing data shared between the main thread and the
   thumbnail thread, i.e. the thumbnail_thread_is_running flag and the
   thumbnails_to_make list. */
static pthread_mutex_t thumbnails_mutex = PTHREAD_MUTEX_INITIALIZER;
/* The currently thumbnailed icon. it also exists in the thumbnails_to_make list
 * to avoid adding it again. Lock thumbnails_mutex when accessing this. */
static KibaThumbnailInfo *currently_thumbnailing = NULL;
/* The list of NautilusThumbnailInfo structs containing information about the
   thumbnails we are making. Lock thumbnails_mutex when accessing this. */
static volatile GList *thumbnails_to_make = NULL;
/* A flag to indicate whether a thumbnail thread is running, so we don't
   start more than one. Lock thumbnails_mutex when accessing this. */
static volatile gboolean thumbnail_thread_is_running = FALSE;

/* The list of NautilusThumbnailInfo structs containing information about the
   thumbnails we are making. Lock thumbnails_mutex when accessing this. */
static volatile GList *thumbnails_to_load = NULL;

static gint thumbnail_load_timeout_id = 0;

static void update_stack_icon_surface (KibaThumbnailInfo *info) {
	KibaStack *stack = info->icon->stack;
	GtkContainer *icon_view;
	GList *l;
	
	info->icon->type = KIBA_STACK_ICON_TYPE_THUMB_FILE;
	g_free (info->icon->mime_icon);
	info->icon->mime_icon = gnome_thumbnail_factory_lookup (info->thumb_factory,
					info->image_uri, info->original_file_mtime);
	cairo_surface_destroy (info->icon->surface->surface);
	info->icon->surface->surface = NULL;
	kiba_stack_icon_update_surface (info->icon);
	
	if (!KIBA_IS_ICON_VIEW_WIN(stack->stack_plugin->view_win))
		return;
	
	icon_view = GTK_CONTAINER(stack->stack_plugin->view_win->icon_view);
	
	for (l = gtk_container_get_children(icon_view); l != NULL; l = l->next) {
		GtkWidget *box = l->data;
		GList *li;
		
		if (g_object_get_data (G_OBJECT(box), "stack-icon") != info->icon)
			continue;
		
		for (li = gtk_container_get_children(GTK_CONTAINER(box)); li != NULL; li = li->next) {
			KibaImage *image = li->data;
			
			if (!KIBA_IS_IMAGE(image))
				continue;
			
			cairo_surface_destroy (image->surface);
			image->surface = NULL;
			
			kiba_image_set_surface (image, info->icon->surface->surface,
					info->icon->surface->width,
					info->icon->surface->height);
		}
	}
}

static gint load_thumbnails_to_icon (gpointer data) {
	GList *l;
	
	for (l = (GList*) thumbnails_to_load; l != NULL; l = (GList*) thumbnails_to_load) {
		KibaThumbnailInfo *info = l->data;
		GList *li;
		
		if (KIBA_IS_STACK_ICON(info->icon)) {
			update_stack_icon_surface (info);
		}
		
		pthread_mutex_lock (&thumbnails_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		thumbnails_to_load = g_list_remove ((GList*) thumbnails_to_load, info);
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&thumbnails_mutex);
		
		g_object_unref (info->thumb_factory);
		g_free (info->image_uri);
		g_free (info);
	}
	
	if (thumbnails_to_make != NULL)
		return TRUE;
	
	thumbnail_load_timeout_id = 0;
	return FALSE;
}

/* thumbnail_thread is invoked as a separate thread to make thumbnails. */
/* note: ported from nautilus */
static gpointer
thumbnail_thread_start (gpointer data)
{
	KibaThumbnailInfo *info = NULL;
	GdkPixbuf *pixbuf;
	time_t current_orig_mtime = 0;
	time_t current_time;
	KibaIconView *icon_view;
	GList *l;

	/* We loop until there are no more thumbails to make, at which point
	   we exit the thread. */
	for (;;) {
		pthread_mutex_lock (&thumbnails_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		
		/* Pop the last thumbnail we just made off the head of the
		   list and free it. I did this here so we only have to lock
		   the mutex once per thumbnail, rather than once before
		   creating it and once after.
		   Don't pop the thumbnail off the queue if the original file
		   mtime of the request changed. Then we need to redo the thumbnail.
		*/
		if (currently_thumbnailing &&
		    currently_thumbnailing->original_file_mtime == current_orig_mtime) {
			g_assert (info == currently_thumbnailing);
			thumbnails_to_make = g_list_remove ((GList*) thumbnails_to_make, currently_thumbnailing);
		}
		currently_thumbnailing = NULL;
		
		/* If there are no more thumbnails to make, reset the
		   thumbnail_thread_is_running flag, unlock the mutex, and
		   exit the thread. */
		if (thumbnails_to_make == NULL) {
			thumbnail_thread_is_running = FALSE;
			pthread_mutex_unlock (&thumbnails_mutex);
			pthread_exit (NULL);
		}
		
		/* Get the next one to make. We leave it on the list until it
		   is created so the main thread doesn't add it again while we
		   are creating it. */
		info = thumbnails_to_make->data;
		currently_thumbnailing = info;
		current_orig_mtime = info->original_file_mtime;
		
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&thumbnails_mutex);
		
		time (&current_time);
		
		/* Don't try to create a thumbnail if the file was modified recently.
		   This prevents constant re-thumbnailing of changing files. */ 
		if (current_time < current_orig_mtime + THUMBNAIL_CREATION_DELAY_SECS &&
		    current_time >= current_orig_mtime) {
 			continue;
		}
		
		/* Create the thumbnail. */
		pixbuf = gnome_thumbnail_factory_generate_thumbnail (info->thumb_factory,
								     info->image_uri,
								     info->mime_type);
		
		if (pixbuf) {
			gnome_thumbnail_factory_save_thumbnail (info->thumb_factory,
								pixbuf,
								info->image_uri,
								current_orig_mtime);
			
			g_object_unref (pixbuf);
		} else {
			gnome_thumbnail_factory_create_failed_thumbnail (info->thumb_factory, 
									 info->image_uri,
									 current_orig_mtime);
		}
		
		pthread_mutex_lock (&thumbnails_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		thumbnails_to_load = g_list_append ((GList*) thumbnails_to_load, info);
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&thumbnails_mutex);
	}
}

static gboolean
get_file_mtime (const char *file_uri, time_t* mtime)
{
	GnomeVFSFileInfo *file_info;
	
	/* gather the info and then compare modification times */
	file_info = gnome_vfs_file_info_new ();
	gnome_vfs_get_file_info (file_uri, file_info, GNOME_VFS_FILE_INFO_FOLLOW_LINKS);

	if (file_info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MTIME)
		*mtime = file_info->mtime;
	else
		*mtime = INVALID_MTIME;
	
	gnome_vfs_file_info_unref (file_info);
	
	return TRUE;
}

void kiba_stack_load_thumb_for_icon (KibaStackIcon *icon) {
	KibaThumbnailInfo *info;
	pthread_attr_t thread_attributes;
	pthread_t thumbnail_thread;
	time_t mtime;
	GnomeVFSURI *vfs_uri;
	char *uri;
	char *escaped_uri;
	GList *l;

	escaped_uri = gnome_vfs_make_uri_from_input (icon->url);
	vfs_uri = gnome_vfs_uri_new (escaped_uri);
	g_free (escaped_uri);
	if (vfs_uri != NULL) {
		uri = gnome_vfs_uri_to_string (vfs_uri, GNOME_VFS_URI_HIDE_NONE);
		gnome_vfs_uri_unref (vfs_uri);
	}
	else {
		uri = g_strdup(icon->url);
	}
	
	/* skip if the file is allready in the thumbnails_to_make list */
	for (l = (GList*) thumbnails_to_make; l != NULL; l = l->next) {
		info = l->data;
		
		if (info->icon == icon || strcmp (uri, info->image_uri) == 0) {
			g_free (uri);
			return;
		}
	}
	
	info = g_new0 (KibaThumbnailInfo, 1);
	info->icon = icon;
	info->image_uri = uri;
	info->mime_type = icon->mime_type;
	info->thumb_factory = g_object_ref (icon->stack->stack_plugin->thumb_factory);
	
	get_file_mtime (info->image_uri, &mtime);
	info->original_file_mtime = mtime;
	
	pthread_mutex_lock (&thumbnails_mutex);
	/*********************************
	 * MUTEX LOCKED
	 *********************************/
	thumbnails_to_make = g_list_append ((GList*) thumbnails_to_make, info);
	/*********************************
	 * MUTEX UNLOCKED
	 *********************************/
	pthread_mutex_unlock (&thumbnails_mutex);
	
	if (!thumbnail_load_timeout_id) {
		thumbnail_load_timeout_id = g_timeout_add (100,
			(GSourceFunc) load_thumbnails_to_icon, icon->stack);
	}
	
	if (thumbnail_thread_is_running)
		return;
	
	/* We create the thread in the detached state, as we don't need/want
	   to join with it at any point. */
	pthread_attr_init (&thread_attributes);
	pthread_attr_setdetachstate (&thread_attributes,
				     PTHREAD_CREATE_DETACHED);
#ifdef _POSIX_THREAD_ATTR_STACKSIZE
	pthread_attr_setstacksize (&thread_attributes, 128*1024);
#endif
	
	/* We set a flag to indicate the thread is running, so we don't create
	   a new one. We don't need to lock a mutex here, as the thumbnail
	   thread isn't running yet. And we know we won't create the thread
	   twice, as we also check thumbnail_thread_starter_id before
	   scheduling this idle function. */
	thumbnail_thread_is_running = TRUE;
	pthread_create (&thumbnail_thread, &thread_attributes,
			thumbnail_thread_start, NULL);

}

/**
Public Functions
**/

void kiba_stack_add (KibaStack     *stack,
		     KibaStackIcon *icon) {
	stack->icons = g_list_append (stack->icons, icon);
	stack->count++;
	
	icon->stack = stack;
	
	if (stack->count == 1) {
		stack->widget = kiba_icon_new("stack");
		gtk_widget_show (stack->widget);
		
		kiba_object_resize (KIBA_OBJECT(stack->widget));
		
		g_signal_connect_after (G_OBJECT (stack->widget), "button-press-event",
				G_CALLBACK (kiba_stack_button_press_cb), stack);
		g_signal_connect_after (G_OBJECT (stack->widget), "button-release-event",
				G_CALLBACK (kiba_stack_button_release_cb), stack);
		g_signal_connect (G_OBJECT (stack->widget), "scroll-event",
				G_CALLBACK (kiba_stack_scroll_cb), stack);
		g_signal_connect (G_OBJECT (stack->widget), "drag-data-received",
				G_CALLBACK (kiba_object_drag_received_cb), stack);
	}
	
	kiba_stack_icon_update_surface (icon);
	if (stack->count == 1)
		kiba_stack_icon_set_state (icon, KIBA_STACK_STATE_ACTIVE);
	
	if (stack->stack_plugin->view_stack == stack &&
	    stack->stack_plugin->view_type == KIBA_STACK_TYPE_CUSTOM) {
		GtkWidget *box;
		box = create_image_box_for_icon(icon, stack->stack_plugin->view_icon_size);
		kiba_icon_view_win_add (stack->stack_plugin->view_win, box);
	}
	
	g_signal_emit (stack, kiba_stack_signals[ICON_ADDED_EVENT], 0, icon);
}

void kiba_stack_remove (KibaStack     *stack,
			KibaStackIcon *icon) {
	GtkWidget *icon_view;
	GtkWidget *box;
	GList *l;
	gboolean first;
	
	g_return_if_fail (KIBA_IS_STACK (stack));
	g_return_if_fail (KIBA_IS_STACK_ICON (icon));
	g_return_if_fail (g_list_find (stack->icons, icon) != NULL);
	
	stack->count--;
	first = (stack->icons->data == icon);
	stack->icons = g_list_remove (stack->icons, icon);
	
	if (stack->stack_plugin->view_stack == stack) {
		KibaIconView *icon_view;
		icon_view = KIBA_ICON_VIEW(stack->stack_plugin->view_win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaStackIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "stack-icon");
			if (check_icon == icon) {
				if (g_list_length(l) == 1) {
					g_signal_handlers_disconnect_by_func (
					     G_OBJECT(stack->widget),
					     G_CALLBACK(kiba_stack_icon_view_size_allocate_cb),
					     stack);
					g_signal_handlers_disconnect_by_func (
					     G_OBJECT(stack->stack_plugin->view_win->win),
					     G_CALLBACK(kiba_stack_icon_view_size_allocate_cb),
					     stack);
				}
				
				kiba_icon_view_win_remove (stack->stack_plugin->view_win, child);
				break;
			}
		}
	}
	
	if (stack->count)
		kiba_object_switch_surface (KIBA_OBJECT(stack->widget));
	
	g_signal_emit (stack,
		kiba_stack_signals[ICON_REMOVED_EVENT],
		0, icon);
	
	/* the removed object was the last one in the stack */
	if (!stack->count) {
		if (stack->uri) {
			g_remove (stack->uri);
		}
		stack->stack_plugin->stacks = g_list_remove (stack->stack_plugin->stacks, stack);
		kiba_stack_destroy (stack);
	}
	else if (first) {
		kiba_stack_icon_set_state (KIBA_STACK_ICON(stack->icons->data),
			KIBA_STACK_STATE_ACTIVE);
	}
}

static void kiba_stack_draw_mime_surface (cairo_t          *cr,
					  cairo_surface_t  *surface,
					  gint              width,
					  gint              height,
					  void             *data) {
	KibaStackMimeSurface *mime_surface = data;
	GdkPixbuf *pixbuf = NULL;
	gint size;
	
	size = MAX(width,height);
	
	kiba_load_icon(g_quark_to_string(mime_surface->icon_quark), size, &pixbuf);
	if (!pixbuf)
		return;
	
	gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
	cairo_paint (cr);
	g_object_unref (pixbuf);
}

cairo_surface_t* kiba_stack_get_mime_surface (KibaStack  *stack,
					      const char *mime_type,
					      const char *mime_icon) {
	cairo_surface_t *ret = NULL;
	cairo_t *cr;
	KibaStackMimeSurface *mime_surface;
	GnomeVFSFileInfo info;
	GList *l;
	GQuark type_quark, icon_quark;
	
	type_quark = g_quark_from_string (mime_type);
	icon_quark = g_quark_from_string (mime_icon);
	
	for (l = stack->stack_plugin->surface_pool; l != NULL; l = l->next) {
		mime_surface = l->data;
		
		if (type_quark == mime_surface->type_quark &&
		    icon_quark == mime_surface->icon_quark) {
			ret = cairo_surface_reference (mime_surface->surface);
			return ret;
		}
	}
	
	mime_surface = g_new0 (KibaStackMimeSurface, 1);
	mime_surface->type_quark = type_quark;
	mime_surface->icon_quark = icon_quark;
	mime_surface->surface = NULL;
	
	stack->stack_plugin->surface_pool = g_list_append (
		stack->stack_plugin->surface_pool, mime_surface);
	
	cr = kiba_win_cairo_create (kiba_win_get_default());
	cairo_update_surface_at_size (cr, &mime_surface->surface,
			stack->widget->allocation.width,
			stack->widget->allocation.height,
			kiba_stack_draw_mime_surface, mime_surface);
	cairo_destroy (cr);
	
	ret = cairo_surface_reference (mime_surface->surface);
	
	return ret;
}

/**
parse the stack text file at stack->uri
and add a KibaStackIcon for each file
**/
static void kiba_stack_parse_file (KibaStack *stack) {
	FILE *file;
	char icon_uri[400];
	
	file = fopen (stack->uri, "r");
	if (file == NULL)
		return;
	
	while (fgets(icon_uri, 400, file) != NULL) {
		KibaStackIcon *stack_icon;
		
		g_strstrip (icon_uri);
		
		if (!g_file_test(icon_uri, G_FILE_TEST_EXISTS))
			continue;
		
		stack_icon = kiba_stack_icon_new(stack);
		if (kiba_stack_icon_set_url (stack_icon, icon_uri))
			kiba_stack_add (stack, stack_icon);
		else
			kiba_stack_icon_destroy (stack_icon);
	}
	
	fclose (file);
}

/**
returns the mime app command or null if no custom mime app is associated with the given mime_type
**/
const char* kiba_stack_get_custom_command_for_mime_type (KibaStackPlugin *plugin,
							 const char *mime_type) {
	GList *l;
	
	for (l = plugin->custom_mime_apps; l != NULL; l = l->next) {
		KibaStackCustomMimeApp *app = l->data;
		
		if (strcmp (app->mime_type, mime_type) == 0)
			return app->command;
	}
	
	return NULL;
}

void kiba_stack_set_custom_mime_app (KibaStackPlugin *plugin,
				     const char *mime_type,
				     const char *command) {
	KibaStackCustomMimeApp *app;
	GList *l;
	
	for (l = plugin->custom_mime_apps; l != NULL; l = l->next) {
		app = l->data;
		
		if (strcmp (app->mime_type, mime_type) == 0) {
			if (strcmp (app->command, command) == 0)
				return;
			else {
				g_free (app->command);
				app->command = g_strdup (command);
				return;
			}
		}
	}
	
	app = g_new0 (KibaStackCustomMimeApp, 1);
	app->command   = g_strdup (command);
	app->mime_type = g_strdup (mime_type);
	
	plugin->custom_mime_apps = g_list_append (plugin->custom_mime_apps, app);
	
	for (l = plugin->stacks; l != NULL; l = l->next) {
		KibaStack *stack = l->data;
		GList *li;
		
		for (li = stack->icons; li != NULL; li = li->next) {
			KibaStackIcon *icon = li->data;
			char *vfs_uri;
			
			if (strcmp (icon->mime_type, app->mime_type) != 0)
				continue;
			
			vfs_uri = gnome_vfs_make_uri_from_shell_arg (icon->url);
			g_free (icon->exec);
			icon->exec = g_strdup_printf ("%s %s", app->command, vfs_uri);
			g_free (vfs_uri);
		}
	}
}

/**
Class Functions
**/

static void kiba_stack_class_init (KibaStackClass *class) {
	kiba_stack_signals[ICON_ADDED_EVENT] =
		g_signal_new ("icon-added",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaStackClass, icon_added),
				NULL, NULL,
				g_cclosure_marshal_VOID__OBJECT,
				G_TYPE_NONE, 1,
				G_TYPE_OBJECT);
	
	kiba_stack_signals[ICON_REMOVED_EVENT] =
		g_signal_new ("icon-removed",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaStackClass, icon_removed),
				NULL, NULL,
				g_cclosure_marshal_VOID__OBJECT,
				G_TYPE_NONE, 1,
				G_TYPE_OBJECT);
}

static void kiba_stack_init (KibaStack *stack) {
	stack->icons  = NULL;
}

KibaStack* kiba_stack_new (KibaStackPlugin *plugin,
			   const char      *uri) {
	KibaStack *stack;
	KibaWin *default_win;
	
	default_win = kiba_win_get_default();
	
	if (!kiba_stack_get_type())
		kiba_stack_register_type (G_TYPE_MODULE(plugin->plugin));
	stack = g_object_new (KIBA_TYPE_STACK, NULL);
	stack->stack_plugin = plugin;
	stack->uri = g_strdup (uri);
	stack->widget = NULL;
	
	kiba_stack_parse_file(stack);
	
	return stack;
}

void kiba_stack_destroy (KibaStack *stack) {
	GList *l;
	
	g_return_if_fail (KIBA_IS_STACK(stack));
	
	if (stack->stack_plugin->view_stack == stack) {
		gtk_widget_destroy (stack->stack_plugin->view_win->win);
	}
	
	for (l = stack->icons; l != NULL; l = stack->icons) {
		KibaStackIcon *icon = l->data;
		
		stack->icons = g_list_remove (stack->icons, l->data);
		
		kiba_stack_icon_destroy (icon);
	}
	g_list_free (stack->icons);
	
	if (KIBA_IS_OBJECT(stack->widget))
		kiba_object_destroy (KIBA_OBJECT(stack->widget));
	
	if (stack->uri) {
		g_free (stack->uri);
	}
	
	g_object_unref (stack);
}
